"""
Basic module utilities
"""
import functools
import keyword
import logging
import os
import pathlib
import warnings
from typing import Callable

from .errors import EnforceError

logger = logging.getLogger(__name__)


HAPPI_SOURCE_PATH = pathlib.Path(__file__).parent


def create_alias(name):
    """
    Clean an alias to be an acceptable Python variable
    """
    return name.replace(' ', '_').replace('.', '_').lower()


def get_happi_entry_value(entry, key, search_extraneous=True):
    extraneous = entry.extraneous
    value = getattr(entry, key, None)
    if value is None and search_extraneous:
        # Try to look at extraneous
        value = extraneous.get(key, None)

    if not value:
        raise ValueError('Invalid Key for Device.')
    return value


def is_number(str_value):
    """
    Checks if it is a valid float number
    """
    try:
        float(str_value)
        return True
    except ValueError:
        return False


def is_a_range(str_value):
    """
    Checks to see if it is a range.
    It needs to have two valid values separated by a comma.
    Ex:
        1,100
        -2,8
        100,2
    """
    if ',' in str_value:
        start, stop = str_value.split(',')
        if is_number(start) and is_number(stop):
            return True
        else:
            logger.error("Possibly provided invalid numbers for a range")
            return False
    else:
        return False


def is_valid_identifier_not_keyword(str_value):
    try:
        if str.isidentifier(str_value) and not keyword.iskeyword(str_value):
            return str_value
    except Exception:
        pass
    raise EnforceError(f'{str_value} is either not a valid Python identifier, '
                       'or is a reserved keyword.')


class OptionalDefault():
    """
    Dummy object to pass to click.prompt if there is no default

    Simply meant to be unique, to avoid collisions with real defaults
    """
    def __str__(self):
        return 'optional'


def optional_enforce(enforce_value):
    def inner(value):
        if isinstance(value, OptionalDefault):
            return value
        else:
            return enforce_value(value)

    return inner


def deprecated(message: str):
    """
    Decorate a method as deprecated with this wrapper.

    Parameters
    ----------
    message : str
        The deprecation warning message.
    """
    def wrapper(func: Callable):
        warned = False

        @functools.wraps(func)
        def warn_and_call(*args, **kwargs):
            nonlocal warned
            if not warned:
                warnings.warn(
                    message=f"{func.__name__} is deprecated: {message}",
                    category=DeprecationWarning,
                )
                warned = True
            return func(*args, **kwargs)

        return warn_and_call

    return wrapper


def build_abs_path(basedir: str, path: str) -> str:
    """
    Builds an abs path starting at basedir if path is not already absolute.
    ~ and ~user constructions will be expanded, so ~/path is considered absolute.
    If path is absolute already, this function returns path without modification.

    Parameters
    ----------
    basedir : str
        If path is not absolute already, build an abspath
        with path starting here.
    path : str
        The path to convert to absolute.
    """
    path = os.path.expanduser(path)
    if not os.path.isabs(path):
        return os.path.abspath(os.path.join(basedir, path))
    return path



import pathlib

from qtpy import QtWidgets

import happi
import happi.qt


class HappiDeviceExplorer(QtWidgets.QFrame):
    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self.filter_label = QtWidgets.QLabel("&Filter")
        self.filter_edit = QtWidgets.QLineEdit()
        self.filter_label.setBuddy(self.filter_edit)

        def set_filter(text):
            self.view.proxy_model.setFilterRegExp(text)

        self.filter_edit.textEdited.connect(set_filter)

        self.view = happi.qt.model.HappiDeviceListView(self)

        self.setLayout(QtWidgets.QGridLayout())
        self.layout().addWidget(self.filter_label, 1, 0)
        self.layout().addWidget(self.filter_edit, 1, 1)
        self.layout().addWidget(self.view, 3, 0, 1, 2)


if __name__ == "__main__":
    app = QtWidgets.QApplication([])
    file_path = pathlib.Path(__file__).resolve()
    db_path = file_path.parent.parent / "db.json"
    cli = happi.Client(path=db_path)

    w = HappiDeviceExplorer()
    w.view.client = cli
    w.view.search(type='OphydItem')
    w.show()

    app.exec_()



import logging
import os
import re
import tempfile
import types
from typing import Any

import pytest

from happi import Client, HappiItem, OphydItem
from happi.backends.json_db import JSONBackend
from happi.errors import DuplicateError, EntryError, SearchError, TransferError

logger = logging.getLogger(__name__)


@pytest.fixture(scope='function')
def xdg_config_home(tmp_path):
    config_home = tmp_path / 'xdg_config_home'
    config_home.mkdir()
    return config_home


@pytest.fixture(scope='function')
def happi_cfg(xdg_config_home):
    # Store current happi config
    xdg_cfg = os.environ.get("XDG_CONFIG_HOME", '')
    happi_cfg = os.environ.get("HAPPI_CFG", '')

    # Setup environment variables
    os.environ['XDG_CONFIG_HOME'] = str(xdg_config_home)
    os.environ['HAPPI_CFG'] = ''

    # Write file
    happi_cfg = xdg_config_home / 'happi.cfg'
    happi_cfg.write_text("""\
[DEFAULT]
backend=json
path=db.json
""")
    yield str(happi_cfg)
    # Restore environment variables
    os.environ["HAPPI_CFG"] = str(happi_cfg)
    os.environ["XDG_CONFIG_HOME"] = xdg_cfg


@pytest.fixture(scope='function')
def happi_cfg_abs(xdg_config_home):
    # Store current happi config
    xdg_cfg = os.environ.get("XDG_CONFIG_HOME", '')
    happi_cfg = os.environ.get("HAPPI_CFG", '')

    # Setup environment variables
    os.environ['XDG_CONFIG_HOME'] = str(xdg_config_home)
    os.environ['HAPPI_CFG'] = ''

    # Write file
    happi_cfg = xdg_config_home / 'happi.cfg'
    happi_cfg.write_text("""\
[DEFAULT]
backend=json
path=/var/run/db.json
""")
    yield str(happi_cfg)
    # Restore environment variables
    os.environ["HAPPI_CFG"] = str(happi_cfg)
    os.environ["XDG_CONFIG_HOME"] = xdg_cfg


def test_find_document(happi_client: Client, item_info: dict[str, Any]):
    doc = happi_client.find_document(**item_info)
    assert doc.pop('prefix') == item_info['prefix']
    assert doc.pop('name') == item_info['name']
    # Remove id and check
    prefix = item_info.pop('prefix')
    doc = happi_client.find_document(**item_info)

    assert doc.pop('prefix') == prefix
    assert doc.pop('name') == item_info['name']
    # Not available
    with pytest.raises(SearchError):
        doc = happi_client.find_document(prefix='Does not Exist')


def test_create_item(happi_client: Client, item_info: dict[str, Any]):
    item = happi_client.create_item(OphydItem, **item_info)
    assert item.prefix == item_info['prefix']
    assert item.name == item_info['name']
    # Invalid Entry
    with pytest.raises(TypeError):
        happi_client.create_item(int)


def test_all_items(happi_client: Client, item: OphydItem):
    assert happi_client.all_items == [item]


def test_add_item(happi_client: Client, valve: OphydItem):
    happi_client.add_item(valve)
    # No duplicates
    with pytest.raises(DuplicateError):
        happi_client.add_item(valve)
    # No incompletes
    d = OphydItem()
    with pytest.raises(EntryError):
        happi_client.add_item(d)


def test_add_and_find_item(
    happi_client: Client,
    valve: OphydItem,
    valve_info: dict[str, Any]
):
    happi_client.add_item(valve)
    loaded_item = happi_client.find_item(**valve_info)
    assert loaded_item.prefix == valve.prefix
    assert loaded_item.name == valve.name


def test_find_item(happi_client: Client, item_info: dict[str, Any]):
    item = happi_client.find_item(**item_info)
    assert isinstance(item, OphydItem)
    assert item.prefix == item_info['prefix']
    assert item.name == item_info['name']
    # Test edit and save
    item.stand = 'DG3'
    item.save()
    loaded_item = happi_client.find_item(**item_info)
    assert loaded_item.prefix == item_info['prefix']
    assert loaded_item.name == item_info['name']
    # Bad load
    bad = {'a': 'b'}
    happi_client.backend.save('a', bad, insert=True)
    with pytest.raises(EntryError):
        happi_client.find_item(**bad)


def test_change_item_name(happi_client: Client, item_info: dict[str, Any]):
    item = happi_client.find_item(**item_info)
    assert item.name != 'new_name'
    item.name = 'new_name'
    item.save()
    # old entry should not exist anymore
    with pytest.raises(SearchError):
        happi_client.find_item(**item_info)
    # new entry with new name should be there
    new_item = happi_client.find_item(name=item.name)
    # prefix or other attributes should be the same
    assert new_item.prefix == item.prefix
    # we should only have one deivce now which is the new one
    assert happi_client.all_items == [new_item]


def test_validate(happi_client: Client):
    # No bad items
    assert happi_client.validate() == list()
    # A single bad item
    happi_client.backend.save('_id', {happi_client._id_key: 'bad'},
                              insert=True)
    assert happi_client.validate() == ['bad']


def test_search(
    happi_client: Client,
    valve: OphydItem,
    item_info: dict[str, Any]
):
    happi_client.add_item(valve)
    res = happi_client.search(name=item_info['name'])
    # Single search return
    assert len(res) == 1
    loaded_item = res[0].item
    assert loaded_item.prefix == item_info['prefix']
    assert loaded_item.name == item_info['name']
    # No results
    assert not happi_client.search(name='not')
    # Returned as dict
    res = happi_client.search(**item_info)
    loaded_item = res[0].item
    assert loaded_item['prefix'] == item_info['prefix']
    assert loaded_item['name'] == item_info['name']
    # Search off keyword
    res = happi_client.search(beamline='LCLS')
    assert len(res) == 2


def test_search_range(happi_client: Client, valve: OphydItem):
    happi_client.add_item(valve)
    # Search between two points
    res = happi_client.search_range('z', start=0, end=500)
    assert len(res) == 2
    # Search between two points, nothing found
    res = happi_client.search_range('z', start=10000, end=500000)
    assert not res
    # Search without an endpoint
    res = happi_client.search_range('z', start=0)
    assert len(res) == 2
    # Search invalid range
    with pytest.raises(ValueError):
        happi_client.search_range('z', start=1000, end=5)


def test_search_regex(
    client_with_three_valves: Client
):
    client = client_with_three_valves

    def find(**kwargs):
        return [
            dict(item) for item in
            client.search_regex(**kwargs, flags=re.IGNORECASE)
        ]

    valve1 = dict(client['VALVE1'])
    valve2 = dict(client['VALVE2'])
    valve3 = dict(client['VALVE3'])

    assert find(beamline='LCLS') == [valve1, valve2, valve3]
    assert find(beamline='lcls') == [valve1, valve2, valve3]
    assert find(beamline='nomatch') == []
    assert find(_id=r'VALVE\d') == [valve1, valve2, valve3]
    assert find(_id='VALVE[13]') == [valve1, valve3]
    assert find(prefix='BASE:VGC[23]:PV') == [valve2, valve3]


def test_get_by_id(
    happi_client: Client,
    valve: OphydItem,
    valve_info: dict[str, Any]
):
    happi_client.add_item(valve)
    name = valve_info['name']
    for k, v in valve_info.items():
        assert happi_client[name][k] == valve_info[k]


def test_remove_item(
    happi_client: Client,
    item: OphydItem,
    valve: OphydItem,
    item_info: dict[str, Any]
):
    happi_client.remove_item(item)
    assert list(happi_client.backend.find(item_info)) == []
    # Invalid item
    with pytest.raises(ValueError):
        happi_client.remove_item(5)
    # Valve not in dictionary
    with pytest.raises(SearchError):
        happi_client.remove_item(valve)


def test_export(happi_client: Client, valve: OphydItem):
    # Setup client with both items
    happi_client.add_item(valve)
    fd, temp_path = tempfile.mkstemp()
    happi_client.export(open(temp_path, 'w+'),
                        sep=',',
                        attrs=['name', 'prefix'])
    exp = open(temp_path).read()
    assert "alias,BASE:PV" in exp
    assert "name,BASE:VGC:PV" in exp
    # Cleanup
    os.remove(temp_path)
    os.close(fd)


def test_load_device(happi_client: Client, item: OphydItem):
    device = happi_client.load_device(name=item.name)
    assert isinstance(device, types.SimpleNamespace)
    assert device.hi == 'oh hello'


@pytest.mark.parametrize(
    'item, target',
    [
        ('item2_dev', 'Item1'),
        ('item1_dev', 'Item2'),
        ('valve', 'Item1')
    ],
)
def test_change_container_fail(
    happi_client: Client,
    item: str,
    target: str,
    request: pytest.FixtureRequest
):
    i = request.getfixturevalue(item)
    t = request.getfixturevalue(target)
    with pytest.raises(TransferError):
        happi_client.change_container(i, t)


def test_change_fail_mandatory(happi_client: Client, item2_dev: HappiItem):
    with pytest.raises(TransferError):
        happi_client.change_container(item2_dev, OphydItem)


@pytest.mark.parametrize(
    'item, target', [('item1_dev', 'Item1'), ('item2_dev', 'Item2')],
)
def test_change_container_pass(
    happi_client: Client,
    item: str,
    target: str,
    request: pytest.FixtureRequest
):
    i = request.getfixturevalue(item)
    t = request.getfixturevalue(target)
    kw = happi_client.change_container(i, t)

    for k in kw:
        assert i.post()[k] == kw[k]


def test_find_cfg(happi_cfg: str):
    # Use our config directory
    assert happi_cfg == Client.find_config()
    # Set the path explicitly using HAPPI env variable
    os.environ['HAPPI_CFG'] = happi_cfg
    assert happi_cfg == Client.find_config()


def test_from_cfg(happi_cfg: str):
    # happi_cfg modifies environment variables to make config discoverable
    client = Client.from_config()
    # Internal db path should be constructed relative to the happi cfg dir
    expected_db = os.path.abspath(os.path.join(os.path.dirname(happi_cfg), 'db.json'))
    print(happi_cfg)
    assert isinstance(client.backend, JSONBackend)
    assert client.backend.path == expected_db


def test_from_cfg_abs(happi_cfg_abs: str):
    # happi_cfg modifies environment variables to make config discoverable
    client = Client.from_config()
    assert isinstance(client.backend, JSONBackend)
    # Ensure the json backend is using the db that we gave an absolute path to.
    assert client.backend.path == '/var/run/db.json'


def test_choices_for_field(happi_client: Client):
    name_choices = happi_client.choices_for_field('name')
    assert name_choices == {'alias'}
    prefix_choices = happi_client.choices_for_field('prefix')
    assert prefix_choices == {'BASE:PV'}
    with pytest.raises(SearchError):
        happi_client.choices_for_field('not_a_field')


def test_searchresults(client_with_three_valves: Client):
    valve1 = client_with_three_valves['VALVE1']
    assert valve1['name'] == 'valve1'
    assert valve1['type'] == 'OphydItem'
    assert valve1['z'] == 300
    assert isinstance(valve1.get(), types.SimpleNamespace)


def test_hashable_searchresults(client_with_three_valves: Client):
    client = client_with_three_valves
    valve1 = client['VALVE1']
    valve2 = client['VALVE2']
    valve3 = client['VALVE3']
    results1 = set(client.search_regex(name='valve1|valve2'))
    results2 = set(client.search_regex(name='valve.*'))

    assert valve1 != 'VALVE1'
    assert valve1 != valve1.metadata

    assert len(results1 & results2) == 2
    assert valve1 in (results1 & results2)
    assert valve2 in (results1 & results2)
    assert len(results1 ^ results2) == 1
    assert valve3 in (results1 ^ results2)
    assert valve2 not in (results1 ^ results2)
    assert results2 == (results1 | results2)


def test_client_mapping(
    client_with_three_valves: Client,
    three_valves: dict[str, dict[str, Any]]
):
    client = client_with_three_valves
    assert len(client) == 3
    assert list(dict(client)) == list(three_valves.keys())
    for name in client:
        assert client[name]['name']


def test_cache_context(monkeypatch, client_with_three_valves: Client):
    client = client_with_three_valves
    valve1 = client['VALVE1']
    valve2 = client['VALVE2']
    valve3 = client['VALVE3']

    cache_clear_count = 0
    orig_clear_cache = client.backend.clear_cache

    def clear_cache():
        nonlocal cache_clear_count
        cache_clear_count += 1
        return orig_clear_cache()

    monkeypatch.setattr(client.backend, "clear_cache", clear_cache)
    with client.retain_cache_context():
        results1 = set(client.search_regex(name='valve1|valve2'))
        results2 = set(client.search_regex(name='valve.*'))

    assert cache_clear_count == 1

    assert valve1 != 'VALVE1'
    assert valve1 != valve1.metadata

    assert len(results1 & results2) == 2
    assert valve1 in (results1 & results2)
    assert valve2 in (results1 & results2)
    assert len(results1 ^ results2) == 1
    assert valve3 in (results1 ^ results2)
    assert valve2 not in (results1 ^ results2)
    assert results2 == (results1 | results2)



"""
MongoDB Backend Implementation using the ``PyMongo`` package.
"""
import logging
import re
from typing import Any, Optional, Union

from pymongo import MongoClient
from pymongo.errors import OperationFailure, ServerSelectionTimeoutError

from ..errors import DatabaseError, DuplicateError, SearchError
from .core import ItemMeta, ItemMetaGen, _Backend

try:
    import bson.regex
except ImportError:
    raise ImportError('Optional dependency mongodb not found; mongodb backend'
                      'is not available.  (bson not found)')

logger = logging.getLogger(__name__)


class MongoBackend(_Backend):
    """
    Abstraction for MongoDB backend.

    Parameters
    ----------
    host : str
        Hostname for MongoDB.
    user : str
        Username for MongoDB instance.
    pw : str
        Password for given username.
    db : str
        Database name within the MongoDB instance.
    collection : str
        MongoDB colleciton name
    timeout : float, optional
        Time to wait for connection attempt.
    port : str or int, optional
        Port of the MongoDB instance.  Will be cast to int.  Defaults to 27017
        (mongo default) if no value is supplied
    auth_source : str, optional
        AuthenticationSource for MongoDB instance, defaults to defaultauthdb if
        no value is supplied
    """

    _timeout = 5
    _conn_str = 'mongodb://{user}:{pw}@{host}/{db}'  # String for login

    def __init__(
        self,
        host: str,
        user: str,
        pw: str,
        db: str,
        collection: str,
        timeout: Optional[str] = None,
        port: Union[str, int] = 27017,
        auth_source: Optional[str] = None,
        **kwargs
    ) -> None:
        # Default timeout
        timeout = timeout or self._timeout
        # Format connection string
        if auth_source:
            conn_str = self._conn_str + f'?authSource={auth_source}'
        else:
            conn_str = self._conn_str

        clean_conn_str = conn_str.format(user=user, pw='...', host=host, db=db)
        conn_str = conn_str.format(user=user, pw=pw, host=host, db=db)
        logging.debug('Attempting connection using %s', clean_conn_str)

        self._client = MongoClient(conn_str, port=int(port),
                                   serverSelectionTimeoutMS=timeout)
        self._db = self._client[db]
        # Load collection
        try:
            if collection not in self._db.list_collection_names():
                raise DatabaseError('Unable to locate collection {} '
                                    'in database'.format(collection))
            self._collection = self._db[collection]
        # Unable to view collection names
        except OperationFailure as e:
            raise PermissionError(e)
        # Unable to connect to MongoDB instance
        except ServerSelectionTimeoutError:
            raise DatabaseError('Unable to connect to MongoDB instance, check '
                                'that the server is running on the host and '
                                'port specified at startup')

    @property
    def all_items(self) -> list[ItemMeta]:
        """List of all item sub-dictionaries."""
        return list(self._collection.find())

    def find(self, to_match: dict[str, Any]) -> ItemMetaGen:
        """
        Yield all instances that match the given search criteria.

        Parameters
        ----------
        to_match : dict
            Requested information.
        """

        yield from self._collection.find(to_match)

    def find_range(
        self,
        key: str,
        *,
        start: Union[int, float],
        stop: Optional[Union[int, float]] = None,
        to_match: dict[str, Any]
    ) -> ItemMetaGen:
        """
        Find an instance or instances that matches the search criteria, such
        that ``start <= entry[key] < stop``.

        Parameters
        ----------
        key : str
            The database key to search.
        start : int or float
            Inclusive minimum value to filter ``key`` on.
        end : float, optional
            Exclusive maximum value to filter ``key`` on.
        to_match : dict
            Requested information, where the values must match exactly.
        """

        if key in to_match:
            raise ValueError('Cannot specify the same key in `to_match` as '
                             'the key for the range.')

        match = {key: {'$gte': start}}
        if stop is not None:
            if start >= stop:
                raise ValueError(f"Invalid range: {start} >= {stop}")
            match[key]['$lt'] = stop

        match.update(**to_match)
        yield from self._collection.find(match)

    def get_by_id(self, _id: str) -> ItemMeta:
        """
        Get an item by ID if it exists, or return None.

        Parameters
        ----------
        _id : str
            The item ID.
        """
        for item in self._collection.find({'_id': _id}):
            return item

    def find_regex(
        self,
        to_match: dict[str, Any],
        *,
        flags=re.IGNORECASE
    ) -> ItemMetaGen:
        """
        Yield all instances that match the given search criteria.

        Parameters
        ----------
        to_match : dict
            Requested information, with each value being a regular expression.
        """

        regexes = {}
        for key, value in to_match.items():
            try:
                reg = re.compile(value, flags=flags)
                regexes[key] = bson.regex.Regex.from_native(reg)
            except Exception as ex:
                raise ValueError(f'Failed to create regular expression from '
                                 f'{key}={value!r}: {ex}') from ex

        yield from self._collection.find(regexes)

    def save(
        self,
        _id: str,
        post: dict[str, Any],
        insert: bool = True
    ) -> None:
        """
        Save information to the database.

        Parameters
        ----------
        _id : str
            ID of item.
        post : dict
            Information to place in database.
        insert : bool, optional
            Whether or not this a new item to the database.

        Raises
        ------
        DuplicateError
            If ``insert`` is `True`, but there is already an item with the
            provided ``_id``.
        SearchError
            If ``insert`` is `False`, but there is no item with the provided
            ``_id``.
        PermissionError
            If the write operation fails due to issues with permissions.
        """

        try:
            # Add to database
            result = self._collection.update_one({'_id': _id},
                                                 {'$set': post},
                                                 upsert=insert)
        except OperationFailure:
            raise PermissionError(
                "Unauthorized command, make sure you are "
                "using a user with write permissions"
            )

        if insert and not result.upserted_id:
            raise DuplicateError(
                "Item with id {} has already been entered "
                "into the database, use load_item and "
                "save if you wish to make changes to the "
                "item".format(_id)
            )

        if not insert and result.matched_count == 0:
            raise SearchError(
                "No item found with id {} please, if this is "
                "a new item, try add_item. If not, make "
                "sure that the item information being sent is "
                "correct".format(_id)
            )

    def delete(self, _id: str) -> None:
        """
        Delete an item instance from the database.

        Parameters
        ----------
        _id : str
            ID of item.
        """

        res = self._collection.delete_one({'_id': _id})
        if res.deleted_count < 1:
            raise SearchError(f'ID not found in database: {_id!r}')



#!/usr/bin/env python3
#
# Happi documentation build configuration file, created by sphinx-quickstart on
# Thu Jan 26 10:44:17 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import os
import sys

import sphinx_rtd_theme  # noqa

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#

module_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "../../")
sys.path.insert(0, module_path)

import happi  # isort: skip # noqa: E402


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinxcontrib.jquery",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosummary",
    # sphinx_click: for 'click' directive (dep sphinx-click)
    "sphinx_click",
    "IPython.sphinxext.ipython_directive",
    "IPython.sphinxext.ipython_console_highlighting",
]

# In case we don't have the backend libraries installed
# we can still build the docs
autodoc_mock_imports = ["psdm_qs_cli", "pymongo"]

autosummary_generate = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "Happi"
copyright = "2021, SLAC National Accelerator Laboratory"
author = "SLAC National Accelerator Laboratory"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = str(happi.__version__)
# The full version, including alpha/beta/rc tags.
release = str(happi.__version__)

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["upcoming_release_notes/template*"]

# The reST default role (used for this markup: `text`)
default_role = "any"

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "Happidoc"


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        "Happi.tex",
        "Happi Documentation",
        "SLAC National Accelerator Laboratory",
        "manual",
    ),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "happi", "Happi Documentation", [author], 1)]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "Happi",
        "Happi Documentation",
        author,
        "Happi",
        "Happi documentation",
        "Miscellaneous",
    ),
]

intersphinx_mapping = {
    "ophyd": ("https://blueskyproject.io/ophyd", None),
    "python": ("https://docs.python.org/3", None),
}



"""
Widget classes designed for atef-to-happi interaction.
"""

from __future__ import annotations

import logging
from typing import Any, ClassVar, Dict, List, Optional, Union

from qtpy import QtCore, QtGui, QtWidgets
from qtpy.QtWidgets import QWidget

import happi
from happi.qt.model import (HappiDeviceListView, HappiDeviceTreeView,
                            HappiViewMixin)

from .designer import DesignerDisplay
from .helpers import ThreadWorker, copy_to_clipboard

logger = logging.getLogger(__name__)


class HappiSearchWidget(DesignerDisplay, QWidget):
    """
    Happi item (device) search widget.

    This widget includes a list view and a tree view for showing all items
    in happi.

    It provides the following signals:
    * ``happi_items_selected`` - one or more happi items were selected.
    * ``happi_items_chosen`` - one or more happi items were chosen by the user.

    To configure multi-item selection, external configuration of
    ``happi_list_view`` and ``happi_tree_view`` are currently required.

    Parameters
    ----------
    parent : QWidget, optional
        The parent widget.

    client : happi.Client, optional
        Happi client instance.  May be supplied at initialization time or
        later.
    """
    filename: ClassVar[str] = 'happi_search_widget.ui'
    happi_items_selected: ClassVar[QtCore.Signal] = QtCore.Signal(
        "QStringList"
    )
    happi_items_chosen: ClassVar[QtCore.Signal] = QtCore.Signal(
        "QStringList"
    )

    _client: Optional[happi.client.Client]
    _last_selected: List[str]
    _search_thread: Optional[ThreadWorker]
    _tree_current_category: str
    _tree_updated: bool
    button_choose: QtWidgets.QPushButton
    button_refresh: QtWidgets.QPushButton
    combo_by_category: QtWidgets.QComboBox
    device_selection_group: QtWidgets.QGroupBox
    edit_filter: QtWidgets.QLineEdit
    happi_list_view: HappiDeviceListView
    happi_tree_view: HappiDeviceTreeView
    label_filter: QtWidgets.QLabel
    layout_by_name: QtWidgets.QHBoxLayout
    list_or_tree_frame: QtWidgets.QFrame
    radio_by_category: QtWidgets.QRadioButton
    radio_by_name: QtWidgets.QRadioButton

    def __init__(
        self,
        parent: Optional[QWidget] = None,
        *,
        client: Optional[happi.Client] = None,
    ):
        super().__init__(parent=parent)
        self._client = None
        self._last_selected = []
        self._tree_current_category = "beamline"
        self._search_thread = None
        self._tree_has_data = False
        self._setup_ui()
        # Set the client at the end, as this may trigger an update:
        self.client = client

    def _setup_ui(self):
        """Configure UI elements at init time."""
        self._setup_tree_view()
        self._setup_list_view()

        def record_selected_items(items: List[str]):
            self._last_selected = items

        self.happi_items_selected.connect(record_selected_items)

        def items_chosen():
            self.happi_items_chosen.emit(list(self._last_selected))

        self.button_refresh.clicked.connect(self.refresh_happi)
        self.button_choose.clicked.connect(items_chosen)
        self.list_or_tree_frame.layout().insertWidget(0, self.happi_list_view)

        self.radio_by_name.clicked.connect(self._select_device_widget)
        self.radio_by_category.clicked.connect(self._select_device_widget)
        self.combo_by_category.currentTextChanged.connect(
            self._category_changed
        )
        self.button_refresh.clicked.emit()

    def _setup_list_view(self):
        """Set up the happi_list_view."""
        def list_selection_changed(
            selected: QtCore.QItemSelection,
            deselected: QtCore.QItemSelection
        ):
            self.happi_items_selected.emit(
                [idx.data() for idx in selected.indexes()]
            )

        view = self.happi_list_view
        view.selectionModel().selectionChanged.connect(
            list_selection_changed
        )

        def item_double_clicked(index: QtCore.QModelIndex):
            self.happi_items_chosen.emit([index.data()])

        view.doubleClicked.connect(item_double_clicked)

        view.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        view.customContextMenuRequested.connect(
            self._list_view_context_menu
        )

        self.edit_filter.textEdited.connect(self._update_filter)

    def _setup_tree_view(self):
        """Set up the happi_tree_view."""
        view = self.happi_tree_view
        view.setVisible(False)
        view.groups = [
            self.combo_by_category.itemText(idx)
            for idx in range(self.combo_by_category.count())
        ]
        self.list_or_tree_frame.layout().insertWidget(0, view)

        def tree_selection_changed(
            selected: QtCore.QItemSelection, deselected: QtCore.QItemSelection
        ):
            items = [
                idx.data() for idx in selected.indexes()
                if idx.parent().data() is not None  # skip top-level items
            ]
            self.happi_items_selected.emit(items)

        view.selectionModel().selectionChanged.connect(
            tree_selection_changed
        )

        view.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        view.customContextMenuRequested.connect(self._tree_view_context_menu)

        self.edit_filter.textEdited.connect(self._update_filter)
        view.proxy_model.setRecursiveFilteringEnabled(True)

    def _update_filter(self, text: Optional[str] = None) -> None:
        """
        Update the list/tree view filters based on the ``edit_filter`` text.
        """
        if text is None:
            text = self.edit_filter.text()

        text = text.strip()
        self.happi_list_view.proxy_model.setFilterRegExp(text)
        self.happi_tree_view.proxy_model.setFilterRegExp(text)

    def _tree_view_context_menu(self, pos: QtCore.QPoint) -> None:
        """Context menu for the happi tree view."""
        self.menu = QtWidgets.QMenu(self)
        index: QtCore.QModelIndex = self.happi_tree_view.indexAt(pos)
        if index is not None:
            def copy(*_):
                copy_to_clipboard(index.data())

            copy_action = self.menu.addAction(f"&Copy: {index.data()}")
            copy_action.triggered.connect(copy)

        self.menu.exec_(self.happi_tree_view.mapToGlobal(pos))

    def _list_view_context_menu(self, pos: QtCore.QPoint) -> None:
        """Context menu for the happi list view."""
        self.menu = QtWidgets.QMenu(self)
        index: QtCore.QModelIndex = self.happi_list_view.indexAt(pos)
        if index is not None:
            def copy(*_):
                copy_to_clipboard(index.data())

            copy_action = self.menu.addAction(f"&Copy: {index.data()}")
            copy_action.triggered.connect(copy)

        self.menu.exec_(self.happi_list_view.mapToGlobal(pos))

    @property
    def selected_device_widget(self) -> Union[
        HappiDeviceListView, HappiDeviceTreeView
    ]:
        """The selected device widget - either the list or tree view."""
        if self.radio_by_name.isChecked():
            return self.happi_list_view

        return self.happi_tree_view

    @QtCore.Slot(str)
    def _category_changed(self, category: str):
        """By-category category has changed."""
        if self._tree_has_data and self._tree_current_category == category:
            return

        self._tree_current_category = category
        self.happi_tree_view.group_by(category)
        # Bugfix (?) otherwise this ends up in descending order
        self.happi_tree_view.model().sort(0, QtCore.Qt.AscendingOrder)
        self.radio_by_category.setChecked(True)
        self._select_device_widget()

    @QtCore.Slot()
    def _select_device_widget(self):
        """Switch between the list/table view."""
        selected = self.selected_device_widget
        for widget in (self.happi_tree_view, self.happi_list_view):
            widget.setVisible(selected is widget)

        if self.happi_tree_view.isVisible() and not self._tree_has_data:
            self._tree_has_data = True
            self.refresh_happi()

    @QtCore.Slot()
    def refresh_happi(self):
        """Search happi again and update the widgets."""
        def search():
            # TODO/upstream: this is coupled with 'search' in the view
            HappiViewMixin.search(self.selected_device_widget)

        def update_gui():
            # TODO/upstream: this is coupled with 'search' in the view
            self.selected_device_widget._update_data()
            self.button_refresh.setEnabled(True)
            self._update_filter()

        def report_error(ex: Exception):
            logger.warning(
                "Failed to update happi information: %s",
                ex, exc_info=ex
            )
            self.button_refresh.setEnabled(True)

        if self._client is None:
            return
        if self._search_thread is not None and self._search_thread.isRunning():
            return

        self.button_refresh.setEnabled(False)
        self._search_thread = ThreadWorker(search)
        self._search_thread.finished.connect(update_gui)
        self._search_thread.error_raised.connect(report_error)
        self._search_thread.start()

    @property
    def client(self) -> Optional[happi.Client]:
        """The client to use for search."""
        return self._client

    @client.setter
    def client(self, client: Optional[happi.Client]):
        self._client = client
        self.happi_tree_view.client = client
        self.happi_list_view.client = client
        self.refresh_happi()


class HappiItemMetadataView(DesignerDisplay, QtWidgets.QWidget):
    """
    Happi item (device) metadata information widget.

    This widget contains a table that displays key and value information
    as provided from the happi client.

    The default context menu allows for copying of keys or values.

    It emits an ``updated_metadata(item_name: str, md: dict)`` when the
    underlying model is updated.

    Parameters
    ----------
    parent : QWidget, optional
        The parent widget.

    client : happi.Client, optional
        Happi client instance.  May be supplied at initialization time or
        later.
    """
    filename: ClassVar[str] = 'happi_metadata_view.ui'
    updated_metadata: ClassVar[QtCore.Signal] = QtCore.Signal(str, object)

    _client: Optional[happi.client.Client]
    _item_name: Optional[str]
    item: Optional[happi.HappiItem]
    label_title: QtWidgets.QLabel
    model: QtGui.QStandardItemModel
    proxy_model: QtCore.QSortFilterProxyModel
    table_view: QtWidgets.QTableView
    _metadata: Dict[str, Any]

    def __init__(
        self,
        parent: Optional[QWidget] = None,
        *,
        client: Optional[happi.Client] = None,
        item_name: Optional[str] = None,
    ):
        super().__init__(parent=parent)
        self._client = None
        self._item_name = None
        self._item = None
        self._setup_ui()
        # Set the client/item at the end, as this may trigger an update:
        self.client = client
        self.item_name = item_name

    def _setup_ui(self):
        """Configure UI elements at init time."""
        self.model = QtGui.QStandardItemModel()

        self.proxy_model = QtCore.QSortFilterProxyModel()
        self.proxy_model.setFilterKeyColumn(-1)
        self.proxy_model.setDynamicSortFilter(True)
        self.proxy_model.setSourceModel(self.model)
        self.table_view.setModel(self.proxy_model)

        self.table_view.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.table_view.customContextMenuRequested.connect(
            self._table_context_menu
        )

    def _table_context_menu(self, pos: QtCore.QPoint) -> None:
        """Context menu when the key/value table is right-clicked."""
        self.menu = QtWidgets.QMenu(self)
        index: QtCore.QModelIndex = self.table_view.indexAt(pos)
        if index is not None:
            def copy(*_):
                copy_to_clipboard(index.data())

            copy_action = self.menu.addAction(f"&Copy: {index.data()}")
            copy_action.triggered.connect(copy)

        self.menu.exec_(self.table_view.mapToGlobal(pos))

    def _update_metadata(self):
        """
        Update the metadata based on ``self.item_name`` using the configured
        client.
        """
        if self.client is None or self.item_name is None:
            return

        try:
            self.item = self.client[self.item_name]
        except KeyError:
            self.item = None

        metadata = dict(self.item or {})
        self._metadata = metadata
        self.updated_metadata.emit(self.item_name, metadata)
        self.model.clear()
        if self.item is None:
            self.label_title.setText("")
            return

        self.label_title.setText(metadata["name"])
        self.model.setHorizontalHeaderLabels(["Key", "Value"])
        skip_keys = {"_id", "name"}
        for key, value in sorted(metadata.items()):
            if key in skip_keys:
                continue

            key_item = QtGui.QStandardItem(str(key))
            value_item = QtGui.QStandardItem(str(value))
            key_item.setFlags(key_item.flags() & ~QtCore.Qt.ItemIsEditable)
            value_item.setFlags(value_item.flags() & ~QtCore.Qt.ItemIsEditable)
            self.model.appendRow([key_item, value_item])

    @property
    def client(self) -> Optional[happi.Client]:
        """The client to use for search."""
        return self._client

    @client.setter
    def client(self, client: Optional[happi.Client]):
        self._client = client
        self._update_metadata()

    @property
    def item_name(self) -> Optional[str]:
        """The item name to search for metadata."""
        return self._item_name

    @item_name.setter
    def item_name(self, item_name: Optional[str]):
        self._item_name = item_name
        self._update_metadata()

    @property
    def metadata(self) -> Dict[str, Any]:
        """The current happi item metadata, as a dictionary."""
        return dict(self._metadata)



import os
import os.path
import tempfile
from typing import Any

import pytest
import simplejson

from happi import Client
from happi.backends.json_db import JSONBackend
from happi.backends.multi_db import MultiBackend
from happi.errors import DuplicateError, SearchError
from happi.loader import load_devices

from .conftest import (requires_mongo, requires_pcdsdevices, requires_py39,
                       requires_questionnaire)


@pytest.fixture(scope='function')
def mockmongo(mockmongoclient):
    return mockmongoclient.backend


@pytest.fixture(scope='function')
def mockjson(item_info: dict[str, Any], valve_info: dict[str, Any]):
    # Write underlying database
    with tempfile.NamedTemporaryFile(mode='w') as handle:
        simplejson.dump({item_info['_id']: item_info},
                        handle)
        handle.flush()
        # Return handle name
        yield JSONBackend(handle.name)


@pytest.fixture(scope='function')
def mockmulti(mockjson, mockmongo, valve_info, item_info):
    """ Create a multi-backend database with a mongo and json backend """
    # modify json backend to have group = 'JSON', mongo to have group='MONGO'
    for backend, group_name in zip([mockjson, mockmongo], ['JSON', 'MONGO']):
        all_docs = backend.all_items
        for doc in all_docs:
            backend.delete(doc[Client._id_key])
            doc['group'] = group_name
            backend.save(doc[Client._id_key], doc, insert=True)

    # add extra device to json backend
    mockjson.save(valve_info[Client._id_key], valve_info, insert=True)

    # add extra device to mongo backend
    extra_info = item_info.copy()
    extra_info['name'] = 'mongo_alias'
    extra_info['_id'] = 'mongo_alias'
    mockmongo.save(extra_info[Client._id_key], extra_info, insert=True)

    # json takes priority over mongo
    return MultiBackend(backends=[mockjson, mockmongo])


@requires_mongo
def test_mongo_find(
    valve_info: dict[str, Any],
    item_info: dict[str, Any],
    mockmongo
):
    mm = mockmongo
    mm._collection.insert_one(valve_info)

    def find(**kwargs):
        return list(mm.find(kwargs))

    assert find(beamline='BLERG') == []
    # Single item by id
    assert [item_info] == find(_id=item_info['_id'])
    # Single item by kwarg
    assert [valve_info] == find(prefix=valve_info['prefix'])
    # No multiple items expected
    assert find(beamline='BLERG') == []
    # Multiple items by id
    assert [item_info] == find(_id=item_info['_id'])
    # Multiple items by kwarg
    assert [item_info] == find(prefix=item_info['prefix'])
    # Multiple items expected
    assert all(info in find(beamline='LCLS')
               for info in (item_info, valve_info))


@requires_mongo
def test_mongo_save(
    mockmongo,
    item_info: dict[str, Any],
    valve_info: dict[str, Any]
):
    # Duplicate item
    with pytest.raises(DuplicateError):
        mockmongo.save(item_info[Client._id_key], item_info, insert=True)

    # Item not found
    with pytest.raises(SearchError):
        mockmongo.save(valve_info[Client._id_key], valve_info, insert=False)

    # Add to database
    mockmongo.save(valve_info[Client._id_key], valve_info, insert=True)
    assert mockmongo._collection.find_one(valve_info) == valve_info


@requires_mongo
def test_mongo_delete(mockmongo, item_info: dict[str, Any]):
    mockmongo.delete(item_info[Client._id_key])
    assert mockmongo._collection.find_one(item_info) is None


def test_json_find(
    valve_info: dict[str, Any],
    item_info: dict[str, Any],
    mockjson
):
    mm = mockjson
    # Write underlying database
    with open(mm.path, 'w+') as handle:
        simplejson.dump({valve_info['_id']: valve_info,
                         item_info['_id']: item_info},
                        handle)

    def find(**kwargs):
        return list(mm.find(kwargs))

    # No single item expected
    assert find(beamline='BLERG') == []
    # Single item by id
    assert [item_info] == find(_id=item_info['_id'])
    # Single item by kwarg
    assert [valve_info] == find(prefix=valve_info['prefix'])
    # No multiple items expected
    assert find(beamline='BLERG') == []
    # Multiple items by id
    assert [item_info] == find(_id=item_info['_id'])
    # Multiple items by kwarg
    assert [item_info] == find(prefix=item_info['prefix'])
    # Multiple items expected
    assert all(info in find(beamline='LCLS')
               for info in (item_info, valve_info))


def test_find_regex(client_with_three_valves, three_valves):
    client = client_with_three_valves

    def find(**kwargs):
        return list(client.backend.find_regex(kwargs))

    valve1 = three_valves['VALVE1']
    valve2 = three_valves['VALVE2']
    valve3 = three_valves['VALVE3']
    assert find(beamline='LCLS') == [valve1, valve2, valve3]
    assert find(beamline='lcls') == [valve1, valve2, valve3]
    assert find(beamline='nomatch') == []
    assert find(_id=r'VALVE\d') == [valve1, valve2, valve3]
    assert find(_id='VALVE[13]') == [valve1, valve3]
    assert find(prefix='BASE:VGC[23]:PV') == [valve2, valve3]


def test_json_delete(mockjson, item_info: dict[str, Any]):
    mockjson.delete(item_info[Client._id_key])
    assert item_info not in mockjson.all_items


def test_json_save(mockjson, item_info: dict[str, Any], valve_info):
    # Duplicate item
    with pytest.raises(DuplicateError):
        mockjson.save(item_info[Client._id_key], item_info, insert=True)

    # Item not found
    with pytest.raises(SearchError):
        mockjson.save(valve_info[Client._id_key], valve_info, insert=False)

    # Add to database
    mockjson.save(valve_info[Client._id_key], valve_info, insert=True)
    assert valve_info in mockjson.all_items


def test_json_initialize():
    jb = JSONBackend("testing.json", initialize=True)
    # Check that the file was made
    assert os.path.exists("testing.json")
    # Check it is a valid json file
    assert jb.load() == {}
    # Check that we can not overwrite the database
    with pytest.raises(PermissionError):
        JSONBackend("testing.json", initialize=True)
    # Cleanup
    os.remove("testing.json")


def test_json_tempfile_location():
    jb = JSONBackend("testing.json", initialize=False)
    assert os.path.dirname(jb.path) == os.path.dirname(jb._temp_path())


def test_json_tempfile_uniqueness():
    jb = JSONBackend("testing.json", initialize=False)
    tempfiles = []
    for _ in range(100):
        tempfiles.append(jb._temp_path())
    assert len(set(tempfiles)) == len(tempfiles)


def test_json_tempfile_remove(monkeypatch):
    # Set consistent temppath
    jb = JSONBackend("testing.json", initialize=False)
    temp_path = jb._temp_path()
    jb._temp_path = lambda: temp_path

    # Ensure file is created, then throw error through patching
    def shutil_move_patch(*args, **kwargs):
        assert os.path.isfile(os.path.join(os.getcwd(), temp_path))
        raise RuntimeError('Simulated testing error.')

    monkeypatch.setattr('shutil.move', shutil_move_patch)

    # Test, and ensure file is deleted appropriately
    with pytest.raises(RuntimeError):
        jb.initialize()
    assert os.path.exists(temp_path) is False


@requires_questionnaire
def test_qs_find(mockqsbackend):
    assert len(list(mockqsbackend.find(dict(beamline='TST')))) == 14
    assert len(list(mockqsbackend.find(dict(name='sam_r')))) == 1


@requires_questionnaire
@requires_pcdsdevices
def test_qsbackend_with_client(mockqsbackend):
    c = Client(database=mockqsbackend)
    assert len(c.all_items) == 14
    assert all(
        d.__class__.__name__ in {'Trigger', 'Motor', 'Acromag', 'LCLSItem'}
        for d in c.all_items
    )
    item_types = [item.__class__.__name__ for item in c.all_items]
    assert item_types.count('Motor') == 7
    assert item_types.count('Trigger') == 2
    # Acromag: six entries, but one erroneously has the same name
    assert item_types.count('Acromag') == 5


@requires_questionnaire
@requires_pcdsdevices
@requires_py39
def test_qsbackend_with_acromag(mockqsbackend):
    c = Client(database=mockqsbackend)
    d = load_devices(*c.all_items, pprint=False).__dict__
    ai1 = d.get('ai_7')
    ao1 = d.get('ao_6')
    assert ai1.__class__.__name__ == 'EpicsSignalRO'
    assert ao1.__class__.__name__ == 'EpicsSignal'


@requires_questionnaire
@requires_pcdsdevices
@requires_py39
def test_beckoff_axis_device_class(mockqsbackend):
    c = Client(database=mockqsbackend)
    d = load_devices(*c.all_items).__dict__
    vh_y = d.get('vh_y')
    sam_x = d.get('sam_x')
    assert vh_y.__class__.__name__ == 'BeckhoffAxis'
    assert sam_x.__class__.__name__ == 'IMS'


@requires_mongo
def test_multi_backend(mockmulti, item_info, valve_info):
    mm = mockmulti
    assert len(mm.all_items) == 3
    assert len(mm.backends[0].all_items) == 2
    assert len(mm.backends[1].all_items) == 2

    # duplicate name/id in both backends
    assert len(list(mm.backends[0].find({'_id': item_info['_id']})))
    assert len(list(mm.backends[1].find({'_id': item_info['_id']})))
    # JSON backend takes priority
    assert list(mm.find({'_id': item_info['_id']}))[0]['group'] == 'JSON'

    # get_id works similarly
    assert mm.backends[0].get_by_id(item_info['_id'])
    assert mm.backends[1].get_by_id(item_info['_id'])
    assert mm.backends[1].get_by_id(item_info['_id'])['group'] == 'MONGO'

    assert mm.get_by_id(item_info['_id'])['group'] == 'JSON'

    dev = list(mm.find_regex({'_id': item_info['_id']}))[0]
    assert mm.get_by_id(item_info['_id']) == dev

    # extra items in backend
    assert mm.get_by_id(valve_info['_id'])
    assert list(mm.find({'_id': valve_info['_id']}))

    assert mm.get_by_id('mongo_alias')
    assert list(mm.find({'_id': 'mongo_alias'}))



from click import UsageError


class DatabaseError(Exception):
    """Raised when an database intitializes improperly."""
    pass


class EntryError(Exception):
    """Raised when there is an invalid happi entry."""
    pass


class DuplicateError(Exception):
    """Raised when a duplicate item is saved."""
    pass


class ContainerError(Exception):
    """Raised by an improperly setup container."""
    pass


class SearchError(Exception):
    """Raised when no item is found while searching."""
    pass


class EnforceError(ValueError, UsageError):
    """Raised when a value fails enforcement checks."""
    def __init__(self, message):
        self.message = str(message)


class TransferError(ValueError):
    """Raised on error transferring item to new container."""
    def __init__(self, message, key):
        self.key = key
        self.message = str(message)
        super().__init__(self.message)



import collections
import copy
import logging
import os
import sys
from collections import OrderedDict
from typing import Any, Optional

from prettytable import PrettyTable

from .errors import ContainerError, EnforceError
from .utils import is_valid_identifier_not_keyword

logger = logging.getLogger(__name__)

# For back-compat to <py3.7
try:
    from re import Pattern
except ImportError:
    from re import _pattern_type as Pattern


class EntryInfo:
    """
    A piece of information related to a specific device.

    These are entered as class attributes for a given device container. They
    help control the information entered into a device.

    Parameters
    ----------
    doc : str
        A short string to document the device.
    optional : bool, optional
        By default all EntryInfo is optional, but in certain cases you may want
        to demand a particular piece of information upon initialization.
    enforce : type, list, compiled regex, or function, optional
        Specify that all entered information is entered in a specific format.
        This can either by a Python type i.e. int, float e.t.c., a list of
        acceptable values, a compiled regex pattern i.e ``re.compile(...)``,
        or custom handling by passing a function that takes in one argument,
        the value. This function must do one of: return the value back as-is,
        return the value back converted to a corrected form, or raise a
        EnforceError.
    default : optional
        A default value for the trait to have if the user does not specify.
        Keep in mind that this should be the same type as ``enforce`` if you
        are demanding a certain type.
    enforce_doc : str, optional
        A human-readable explanation of the enforce field.  Will be printed
        if the entered information does not follow the ``enforce`` type
    include_default_as_kwarg : bool, optional
        Defaults to True.  If a kwargs EntryInfo sets this to False, all kwargs
        will be compared to their corresponding Entries in the item and
        omitted from the kwargs dictionary if their value matches the Entry's
        default.
        This can also be set on an individual Entry basis.  The setting on
        an individual entry will only be taken into consideration if the
        kwarg EntryInfo has this set to True (default).

    Raises
    ------
    ContainerError
        If there is an error with the way the enforced value interacts with its
        default value, or if the piece of information entered is unenforcable
        based on the the settings.

    Example
    ------
    .. code::

        class MyItem(HappiItem):
            my_field = EntryInfo('My generated field')
            number   = EntryInfo('Device number', enforce=int, default=0,
                                 enforce_doc='This must be a number')
    """

    def __init__(
        self,
        doc: Optional[str] = None,
        optional: Optional[bool] = True,
        enforce: Optional[Any] = None,
        default: Optional[Any] = None,
        enforce_doc: Optional[str] = None,
        include_default_as_kwarg: bool = True
    ):
        self.key = None  # Set later by parent class
        self.doc = doc
        self.enforce = enforce
        self.optional = optional
        self.include_default_as_kwarg = include_default_as_kwarg
        self.enforce_doc = str(enforce_doc or '')

        # Explicitly set default to None b/c this is how we ensure mandatory
        # information was set
        if optional:
            self.default = default
        else:
            self.default = None

        # Check that default value is correct type
        try:
            self.enforce_value(default)
        except ValueError:
            raise ContainerError('Default value must match the enforced type')

    def enforce_value(self, value):
        """
        Enforce the rules of the EntryInfo.

        Accepts a ``value``, verifies that it meets the criteria specified in
        the ``enforce`` attribute, and returns the same value, except that it
        will be converted to the correct type if ``enforce`` is a type.

        Raises
        ------
        EnforceError
            If the value is not the correct type, or does not match the
            pattern.
        """
        if not self.enforce or value is None:
            return value

        elif self.enforce is bool and isinstance(value, str):
            # Special case for booleans, since the value may come in as a
            # string, and bool('False') evaluates to True
            true_values = ['true', 't', 'yes', 'y']
            false_values = ['false', 'f', 'no', 'n']
            if value.lower() in true_values:
                return True
            elif value.lower() in false_values:
                return False
            else:
                raise EnforceError(f'{value} as a string is not interpretable '
                                   f'as a boolean.  {self.enforce_doc}')

        elif callable(self.enforce):
            # Try and convert to type or custom handling
            # Otherwise raise EnforceError for types, custom handling otherwise
            try:
                return self.enforce(value)
            except ValueError as e:
                if self.enforce_doc:
                    raise EnforceError(self.enforce_doc) from e
                else:
                    raise EnforceError(e)

        elif isinstance(self.enforce, (list, tuple, set)):
            # Check that value is in list, otherwise raise EnforceError
            if value not in self.enforce:
                raise EnforceError(f'{value} was not found in the enforce '
                                   f'list {self.enforce}. {self.enforce_doc}')
            return value

        elif isinstance(self.enforce, Pattern):
            # Try and match regex patttern, otherwise raise EnforceError
            try:
                if not self.enforce.match(value):
                    raise EnforceError(
                        f'{self.key}={value!r} did not match the enforced '
                        f'pattern: ({self.enforce.pattern}). '
                        f'{self.enforce_doc}'
                    )
            except TypeError as e:
                raise EnforceError(e)

            return value

        # Invalid enforcement
        else:
            raise ContainerError('EntryInfo {} has an invalid enforce'
                                 ''.format(self.key))

    def make_docstring(self, parent_class):
        if self.doc is not None:
            return self.doc

        doc = [f'{self.__class__.__name__} attribute',
               '::',
               '',
               ]

        doc.append(repr(self))
        doc.append('')
        return '\n'.join(doc)

    def __get__(self, instance, owner):
        if instance is None:
            return self

        return instance.__dict__[self.key]

    def __set__(self, instance, value):
        instance.__dict__[self.key] = self.enforce_value(value)

    def __repr__(self):
        return 'EntryInfo {} (optional={}, default={})'.format(self.key,
                                                               self.optional,
                                                               self.default)

    def __copy__(self):
        return EntryInfo(doc=self.doc, optional=self.optional,
                         enforce=self.enforce, default=self.default,
                         enforce_doc=self.enforce_doc)


class _HappiItemBase:
    def __init_subclass__(cls, **kargs):
        # These attributes are not to be overwritten by subclasses
        RESERVED_ATTRS = [
            '_id', '_info_attrs',
            'creation', 'entry_info', 'info_names', 'last_edit',
            'mandatory_info', 'post', 'save',
        ]

        # Create dict to hold information
        cls._info_attrs = OrderedDict()

        # Handle multiple inheritance
        for base in reversed(cls.mro()[1:]):
            if not hasattr(base, '_info_attrs'):
                continue

            for attr, info in base._info_attrs.items():
                cls._info_attrs[attr] = info

        # Load from highest classEntry
        entries = [(attr, entry)
                   for attr, entry in cls.__dict__.items()
                   if isinstance(entry, EntryInfo)
                   ]

        for attr, entry in entries:
            if attr in RESERVED_ATTRS:
                raise TypeError(
                    f"The attribute name {attr} is used by the HappiItem "
                    f"class and cannot be used as a name for EntryInfo. "
                    f"Choose a different name."
                )

            cls._info_attrs[attr] = entry

        # Notify Info of key names
        for attr, info in cls._info_attrs.items():
            info.key = attr

        # Create docstring information
        for info in cls._info_attrs.values():
            info.__doc__ = info.make_docstring(cls)

        # Store Entry Information
        cls.entry_info = list(cls._info_attrs.values())
        cls.info_names = [info.key for info in cls.entry_info]
        cls.mandatory_info = [info.key for info in cls.entry_info
                              if not info.optional]


class HappiItem(_HappiItemBase, collections.abc.Mapping):
    """
    The smallest description of an object that can be entered in ``happi``.

    The class does not need to be intialized with any specific piece of
    information except a name, but all of the attributes listed by
    :attr:`HappiItem.info_names` can be used to assign values to
    :class:`.EntryInfo` upon initialization.  Pieces of information that are
    deemed mandatory by the class must be filled in before the device is loaded
    into the database. See :attr:`.mandatory_info` to see which
    attributes are neccesary.

    Additional metadata can be given to the device in the form of keywords on
    initialization, this information is kept in the :attr:`.extraneous`
    attribute, and will be saved in to the database as long as it does not
    clash with an existing piece of metadata that the client uses to organize
    devices.

    Attributes
    ----------
    entry_info : list
        A list of all the :class:`.EntryInfo` associated with the device.
    extraneous : dict
        Storage for information supplied during initialization that does not
        correspond to a specific EntryInfo.

    Raises
    ------
    EntryError
        If a piece of information supplied at startup is of the incorrect type.
    ContainerError
        If one of the pieces of :class:`.EntryInfo` has a default value of the
        incorrect type.

    Example
    -------
    .. code ::

        d = HappiItem(name = 'my_device',         # Alias name for device
                      note  = 'Example',          # Piece of arbitrary metadata
                     )
    """

    name = EntryInfo("Shorthand Python-valid name for the Python instance",
                     optional=False,
                     enforce=is_valid_identifier_not_keyword)
    device_class = EntryInfo("Python class that represents the instance",
                             enforce=str)
    args = EntryInfo("Arguments to pass to device_class",
                     enforce=list, default=[])
    kwargs = EntryInfo("Keyword arguments to pass to device_class",
                       enforce=dict, default={})
    active = EntryInfo('Whether the instance is actively deployed',
                       enforce=bool, default=True)
    documentation = EntryInfo("Relevant documentation for the instance",
                              enforce=str)

    def __init__(self, **kwargs):
        # Load given information into device class
        for info in self.entry_info:
            if info.key in kwargs.keys():
                setattr(self, info.key, kwargs.pop(info.key))
            else:
                setattr(self, info.key, info.default)
        # Handle additional information
        self.extraneous = kwargs
        if kwargs:
            logger.debug("Additional information for %s was defined %s",
                         self.name, ', '.join(self.extraneous))

    def show_info(self, handle=sys.stdout):
        """
        Show the device instance information in a PrettyTable.

        Parameters
        ----------
        handle : file-like, optional
            Option to write to a file-like object.
        """

        pt = PrettyTable(['EntryInfo', 'Value'])
        pt.align = 'r'
        pt.align['EntryInfo'] = 'l'
        pt.align['Value'] = 'l'
        pt.float_format = '8.5'

        # Gather all device information, do not show private
        # information that begins with an underscore
        show_info = self.post()
        public_keys = sorted([key for key in show_info.keys()
                              if not key.startswith('_')])
        for key in public_keys:
            pt.add_row([key, show_info[key]])

        # initialize table so we can grab widths
        _ = pt.get_string()

        # set width of Value column dynamically
        # set width(Value) = terminal width - width(EntryInfo column)
        entry_info_width = pt._widths[0] + 10  # account for padding
        try:
            term_width = os.get_terminal_size()[0]
            pt._max_width = {'Value': max(60, term_width - entry_info_width)}
        except OSError:
            # non-interactive mode (piping results). No max width
            pass

        print(pt, file=handle)

    def post(self):
        """
        Create a document to be loaded into the happi database.

        Returns
        -------
        post : dict
            Dictionary of all contained information.
        """

        # Grab all the specified information
        post = {key: getattr(self, key) for key in self.info_names}

        # Add additional metadata
        if self.extraneous:
            post.update(self.extraneous)

        return post

    def __getitem__(self, item):
        return self.post()[item]

    def __iter__(self):
        yield from self.info_names
        if self.extraneous:
            yield from self.extraneous

    def __len__(self):
        return len(self.post())

    def __copy__(self):
        return self.__deepcopy__({})

    def __deepcopy__(self, memo):
        device_info = copy.deepcopy(dict(self))
        return type(self)(**device_info)

    def save(self):
        """Overwritten when the device instance is loaded from the client."""
        raise NotImplementedError

    def __repr__(self):
        return '{} (name={})'.format(self.__class__.__name__,
                                     self.name)

    def __eq__(self, other):
        return (self.post() == other.post())


class OphydItem(HappiItem):
    prefix = EntryInfo("A base PV for all related records",
                       optional=False, enforce=str)

    args = copy.copy(HappiItem.args)
    args.default = ['{{prefix}}']
    kwargs = copy.copy(HappiItem.kwargs)
    kwargs.default = {'name': '{{name}}'}



"""
Helper QObject classes for managing dataclass instances.

Contains utilities for synchronizing dataclass instances between
widgets.
"""
from __future__ import annotations

import functools
import logging
import platform
from typing import Any, Callable, Dict, List, Optional, Tuple

from qtpy import QtCore, QtGui, QtWidgets

logger = logging.getLogger(__name__)


class ThreadWorker(QtCore.QThread):
    """
    Worker thread helper.  For running a function in a background QThread.

    Parameters
    ----------
    func : callable
        The function to call when the thread starts.
    *args
        Arguments for the function call.
    **kwargs
        Keyword arguments for the function call.
    """

    error_raised = QtCore.Signal(Exception)
    returned = QtCore.Signal(object)
    func: Callable
    args: Tuple[Any, ...]
    kwargs: Dict[str, Any]
    return_value: Any

    def __init__(self, func: Callable, *args, **kwargs):
        super().__init__()
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.return_value = None

    @QtCore.Slot()
    def run(self):
        try:
            self.return_value = self.func(*self.args, **self.kwargs)
        except Exception as ex:
            logger.exception(
                "Failed to run %s(*%s, **%r) in thread pool",
                self.func,
                self.args,
                self.kwargs,
            )
            self.return_value = ex
            self.error_raised.emit(ex)
        else:
            self.returned.emit(self.return_value)


def run_in_gui_thread(
    func: Callable,
    *args,
    _start_delay_ms: int = 0,
    **kwargs
):
    """Run the provided function in the GUI thread."""
    QtCore.QTimer.singleShot(
        _start_delay_ms,
        functools.partial(func, *args, **kwargs)
    )


def get_clipboard() -> Optional[QtGui.QClipboard]:
    """Get the clipboard instance. Requires a QApplication."""
    app = QtWidgets.QApplication.instance()
    if app is None:
        return None

    return QtWidgets.QApplication.clipboard()


def get_clipboard_modes() -> List[int]:
    """
    Get the clipboard modes for the current platform.

    Returns
    -------
    list of int
        Qt-specific modes to try for interacting with the clipboard.
    """
    clipboard = get_clipboard()
    if clipboard is None:
        return []

    if platform.system() == "Linux":
        # Mode selection is only valid for X11.
        return [
            QtGui.QClipboard.Selection,
            QtGui.QClipboard.Clipboard
        ]

    return [QtGui.QClipboard.Clipboard]


def copy_to_clipboard(text: str, *, quiet: bool = False):
    """
    Copy ``text`` to the clipboard.

    Parameters
    ----------
    text : str
        The text to copy to the clipboard.

    quiet : bool, optional, keyword-only
        If quiet is set, do not log the copied text.  Defaults to False.
    """
    clipboard = get_clipboard()
    if clipboard is None:
        return None

    for mode in get_clipboard_modes():
        clipboard.setText(text, mode=mode)
        event = QtCore.QEvent(QtCore.QEvent.Clipboard)
        app = QtWidgets.QApplication.instance()
        if app is not None:
            app.sendEvent(clipboard, event)

    if not quiet:
        logger.warning(
            (
                "Copied text to clipboard:\n"
                "-------------------------\n"
                "%s\n"
                "-------------------------\n"
            ),
            text
        )


def get_clipboard_text() -> str:
    """
    Get ``text`` from the clipboard. If unavailable or unset, empty string.

    Returns
    -------
    str
        The clipboard text, if available.
    """
    clipboard = get_clipboard()
    if clipboard is None:
        return ""
    for mode in get_clipboard_modes():
        text = clipboard.text(mode=mode)
        if text:
            return text
    return ""



"""
Base backend database options.
"""
import logging
from collections.abc import Generator
from typing import Any

logger = logging.getLogger(__name__)


ItemMeta = dict[str, Any]
ItemMetaGen = Generator[ItemMeta, None, None]


class _Backend:
    """
    Base class for backend database.

    Backends are internal to happi and are subject to change.  Users should
    interact with the :class:`happi.Client`.
    """

    @property
    def all_items(self) -> list[ItemMeta]:
        """
        List of all items in the database.

        This base class ``_Backend`` provides no implementation and requires
        that a subclass provide its customized implementation.
        """
        raise NotImplementedError

    def clear_cache(self) -> None:
        """
        Request to clear any cached data.

        Optional implementation may be customized in subclass.
        """

    def find(self, multiples: bool = False, **kwargs) -> ItemMetaGen:
        """
        Find an instance or instances that matches the search criteria.

        This base class ``_Backend`` provides no implementation and requires
        that a subclass provide its customized implementation.

        Parameters
        ----------
        multiples : bool
            Find a single result or all results matching the provided
            information.
        kwargs
            Requested information.
        """
        raise NotImplementedError

    def save(
        self,
        _id: str,
        post: dict[str, Any],
        insert: bool = True
    ) -> None:
        """
        Save information to the database.

        This base class ``_Backend`` provides no implementation and requires
        that a subclass provide its customized implementation.

        Parameters
        ----------
        _id : str
            ID of device.
        post : dict
            Information to place in database.
        insert : bool, optional
            Whether or not this a new device to the database.

        Raises
        ------
        DuplicateError
            If ``insert`` is `True`, but there is already a device with the
            provided ``_id``.
        SearchError
            If ``insert`` is `False`, but there is no device with the provided
            ``_id``.
        PermissionError
            If the write operation fails due to issues with permissions.
        """
        raise NotImplementedError

    def delete(self, _id: str) -> None:
        """
        Delete a device instance from the database.

        This base class ``_Backend`` provides no implementation and requires
        that a subclass provide its customized implementation.

        Parameters
        ----------
        _id : str
            ID of device.

        Raises
        ------
        PermissionError
            If the write operation fails due to issues with permissions.
        """
        raise NotImplementedError



"""
This module contains functions and helpers for auditing the happi database.
Checks are simple functions that take a happi.SearchResult a positional
argument and returns None when successful.  When a check fails, it should throw
an Exception with a helpful error message.  These exception messages will be
caught and organized by the cli audit tool.
"""
from __future__ import annotations

import contextlib
import inspect
import io
import sys
from typing import Callable, Optional, TypedDict

from jinja2 import DebugUndefined, Environment, meta

from . import SearchResult

CLIENT_KEYS = ['_id', 'type', 'creation', 'last_edit']


class CheckResults(TypedDict):
    name: list[str]
    success: list[bool]
    check: list[str]
    msg: list[str]


class ItemAuditInfo(TypedDict):
    failed_check: list[str]
    audit_errors: list[str]


class AuditResults(TypedDict):
    test_results: CheckResults
    audited: int
    failures: int
    failed_names: list[str]
    items: dict[str, ItemAuditInfo]


Check = Callable[[SearchResult], None]
CheckList = list[Check]


def check_instantiation(result: SearchResult) -> None:
    """
    Check if the device can be instantiated
    Simply attempts to get the device from the happi.SearchResult

    Subscription callback exceptions are silenced in ophyd, and will
    not cause this check to fail.
    """
    result.get()


def check_extra_info(
    result: SearchResult,
    ignore_keys: Optional[list[str]] = None
) -> None:
    """
    Check if there is any extra info in the result that does not
    have a corresponding EntryInfo on the container.

    Ignores the presence of client-side metadata keys
    - id
    - type
    - creation
    - last_edit
    """
    if ignore_keys is None:
        ignore_keys = CLIENT_KEYS

    extra = result.item.extraneous.copy()
    for key in ignore_keys:
        extra.pop(key, None)
    if len(extra) != 0:
        raise ValueError(f'Un-enforced metadata found: {list(extra.keys())}')


def check_name_match_id(result: SearchResult) -> None:
    """
    Check if the item's ``_id`` field matches its ``name``.
    This is a convention we have held for a while, making searching
    in either direction easier.
    """
    id_field = result.metadata.get('_id')
    name_field = result.metadata.get('name')
    if id_field != name_field:
        raise ValueError(f'id: {id_field} != name: {name_field}')


def check_wait_connection(result: SearchResult) -> None:
    """
    Check if all the signals on the device can connect to their PV's

    This can take a long time, consider filtering your query before running
    """
    dev = result.get()

    if not (hasattr(dev, 'wait_for_connection') and
            callable(getattr(dev, 'wait_for_connection'))):
        raise ValueError('device has no wait_for_connection method, '
                         'and is likely not an ophyd v1 device')

    try:
        dev.wait_for_connection(timeout=5)
    except TimeoutError as te:
        # If we encounter a timeout error, gather some more detailed stats
        sigs = list(sig.item for sig in dev.walk_signals())
        conn_sigs = sum(sig.connected for sig in sigs)

        raise ValueError(f'{conn_sigs}/{len(sigs)} signals connected. \n'
                         f'original error: {te}')


def check_args_kwargs_match(result: SearchResult) -> None:
    """
    Check that arguments that request information from the container
    have information in those entries.

    i.e.: makes sure there is information in the entry named "extra"
          if the kwargs = {'extra': '{{extra}}'}
    """
    # Happi fills in values when search result is created, must pull
    # raw document from client level
    cl = result.client
    doc = cl.find_document(**{cl._id_key: result[cl._id_key]})
    # pick out any jinja-like template
    env = Environment(undefined=DebugUndefined)

    # render template and check if any variables were undefined
    template = env.from_string(str(doc))
    rendered = template.render(**doc)
    undefined = meta.find_undeclared_variables(env.parse(rendered))

    if len(undefined) != 0:
        raise ValueError(f'undefined variables found in document: {undefined}')


def find_unfilled_mandatory_info(
    result: SearchResult
) -> list[str]:
    """
    Return all mandatory fields that are missing a value
    """
    return [info for info in result.item.mandatory_info
            if getattr(result.item, info) is None]


def find_unfilled_optional_info(
    result: SearchResult
) -> list[str]:
    """
    Return all optional fields that are missing a value
    """
    # cannot getattr for extraneous info, shortcircuit conditional
    return [info for info in list(result.item.keys())
            if info not in result.item.mandatory_info
            and info not in result.item.extraneous.keys()
            and getattr(result.item, info) is None]


def check_unfilled_mandatory_info(result: SearchResult) -> None:
    """
    Check that all mandatory EntryInfo have a value.

    This identifies problems originating from a mismatch between
    an item's backend representation and desired container.
    This may arise from manual editing of the backend database.
    """
    unfilled_info = find_unfilled_mandatory_info(result)
    if len(unfilled_info) != 0:
        raise ValueError(f'unfilled mandatory information found: {unfilled_info}')


def verify_result(
    result: SearchResult,
    check: Callable[[SearchResult], None]
) -> tuple[bool, str, str]:
    """
    Validate device against the provided check

    Parameters
    ----------
    result : happi.SearchResult
        result to be verified

    check : Callable[[SearchResult], None]
        a verification function that raise exceptions if
        verification fails.

    Returns
    -------
    bool
        whether or not the device passed the checks
    str
        name of check that failed, if any
    str
        error message describing reason for failure and possible steps
        to fix.  Empty string if validation is successful
    """
    success, msg = True, ""
    # verify checks have the correct signature, as much as is reasonable
    sig = inspect.signature(check)
    sig.bind(result)

    try:
        check(result)
    except Exception as ex:
        success = False
        msg = str(ex)

    return success, check.__name__, msg


def audit(
    results: list[SearchResult],
    redirect: bool = True,
    verbose: bool = False,
    check_list: Optional[CheckList] = None,
    catch_keyboard_interrupt: bool = True,
) -> AuditResults:
    """
    Audit the given ``SearchResult`` items.

    Parameters
    ----------
    results : list[SearchResult]
        The search results from the happi client to audit.
    redirect : bool
        During the audit process, capture standard output and standard error to
        avoid connection and error-related spam.
    verbose : bool
        Output status messages during the audit process.
    check_list : list[Check]
        The callable checks to perform.
    catch_keyboard_interrupt : bool, optional
        Catch and record when the user attempts to cancel the audit process.

    Returns
    -------
    AuditResults
        The summarized results of the audit.
    """
    if check_list is None:
        # If unspecified, use all checks
        check_list = list(checks)

    test_results: CheckResults = {
        "name": [],
        "success": [],
        "check": [],
        "msg": [],
    }

    @contextlib.contextmanager
    def maybe_redirect():
        if redirect:
            f = io.StringIO()
            with contextlib.redirect_stderr(f), contextlib.redirect_stdout(f):
                yield
        else:
            yield

    try:
        for i, res in enumerate(results):
            if verbose and sys.__stdout__.isatty():
                print(f"checking device #: {i}", end="\r")

            # Capture stdout, stderr for this audit
            with maybe_redirect():
                for check_fn in check_list:
                    try:
                        success, check, msg = verify_result(res, check_fn)
                    except KeyboardInterrupt:
                        test_results["name"].append(res.item.name)
                        test_results["success"].append(False)
                        test_results["check"].append("")
                        test_results["msg"].append("Interrupted by user")
                        raise
                    else:
                        test_results["name"].append(res.item.name)
                        test_results["success"].append(success)
                        test_results["check"].append(check)
                        test_results["msg"].append(msg)
    except KeyboardInterrupt:
        if not catch_keyboard_interrupt:
            raise
        if verbose:
            print("Caught KeyboardInterrupt; exiting early...")

    unique_fails = {
        test_results["name"][i]
        for i in range(len(test_results["name"]))
        if not test_results["success"][i]
    }

    audit_results: AuditResults = {
        "test_results": test_results,
        "audited": len(results),
        "failures": len(unique_fails),
        "failed_names": sorted(unique_fails),
        "items": {
            name: {"failed_check": [], "audit_errors": []}
            for name in [res.item.name for res in results]
        },
    }
    item_info = audit_results["items"]
    for name, success, check, msg in zip(
        test_results["name"],
        test_results["success"],
        test_results["check"],
        test_results["msg"],
    ):
        if not success:
            item_info[name]['failed_check'].append(check)
            item_info[name]['audit_errors'].append(msg)

    return audit_results


checks = [check_instantiation, check_extra_info, check_name_match_id,
          check_wait_connection, check_args_kwargs_match,
          check_unfilled_mandatory_info]



"""
Backend implementation for parsing the LCLS Questionnaire.
"""
import functools
import logging
import re
from typing import Optional

from psdm_qs_cli import QuestionnaireClient

from ..errors import DatabaseError
from .json_db import JSONBackend

logger = logging.getLogger(__name__)


class RequiredKeyError(KeyError):
    """Required key not found in questionnaire."""
    ...


def _create_motor_callable(name, beamline, info):
    """Create a motor entry."""
    container = 'pcdsdevices.happi.containers.Motor'
    class_name = None
    kwargs = {'name': '{{name}}'}
    prefix = info['pvbase']
    if info.get('stageidentity') == 'Beckhoff':
        class_name = 'pcdsdevices.device_types.BeckhoffAxis'
    return create_entry(name, beamline, prefix, kwargs, container,
                        info, class_name)


def _create_trig_callable(name, beamline, info):
    """Create a trigger entry."""
    container = 'pcdsdevices.happi.containers.Trigger'
    kwargs = {'name': '{{name}}'}
    prefix = info['pvbase']
    return create_entry(name, beamline, prefix, kwargs, container, info)


def _create_ai_ao_callable(name, beamline, info):
    """Create an acrommag channel entry."""
    container = 'pcdsdevices.happi.containers.Acromag'
    class_name = 'pcdsdevices.device_types.AcromagChannel'
    prefix = info['pvbase']
    ch = info.get('channel')
    if not ch:
        raise RequiredKeyError('Unable to create an acromag input channel '
                               'entry without a channel')
    name_prefix = 'ai_' if ':ai' in prefix else 'ao_'
    name = f'{name_prefix}{ch}'
    kwargs = {'name': '{{name}}', 'channel': ch}
    return create_entry(name, beamline, prefix, kwargs, container,
                        info, class_name)


def create_entry(name, beamline, prefix, kwargs, container,
                 info, class_name=None):
    """
    Create a happi_entry.

    Parameters
    ----------
    name : str
        Item name.
    beamline : str
        The beamline with which to associate the entry.
    prefix : str
        Epics base PV.
    kwargs : dict
        Information to pass through to the device, upon initialization
    class_name : str
        The class name to report in the new entry.
    container : str
        The container name to report in the new entry.
    info : dict
        Device information from `_translate_items`.
    """

    entry = {
        '_id': name,
        'active': True,
        'args': ['{{prefix}}'],
        'beamline': beamline,
        'kwargs': kwargs,
        'lightpath': False,
        'name': name,
        'prefix': prefix,
        'type': container,
        **info,
    }
    if class_name is not None:
        entry['device_class'] = class_name

    return entry


# Map of (questionnaire type) to:
#   1. Device class (or factory function) and
#   2. Happi container


DEFAULT_TRANSLATIONS = {
    'motors': _create_motor_callable,
    'trig': _create_trig_callable,
    'ao': _create_ai_ao_callable,
    'ai': _create_ai_ao_callable,
}


class QuestionnaireHelper:
    def __init__(self, client: QuestionnaireClient):
        self._client = client
        self._experiment = None
        self.experiment_to_proposal = client.getExpName2URAWIProposalIDs()

    def __repr__(self) -> str:
        try:
            return (
                f'<{self.__class__.__name__} experiment={self.experiment} '
                f'run_number={self.run_number} proposal={self.proposal} '
                f'beamline={self.beamline}>'
            )
        except Exception:
            return f'<{self.__class__.__name__} experiment={self.experiment}>'

    @property
    def experiment(self) -> str:
        """The experiment name."""
        return self._experiment

    @experiment.setter
    def experiment(self, experiment: str):
        self._experiment = experiment

        # Proposals are per-experiment: clear the cache.
        self.get_proposal_list.cache_clear()
        self.get_run_details.cache_clear()

    @property
    def proposal(self):
        """Get the proposal number for the configured experiment."""
        if self.experiment is None:
            raise RuntimeError('Experiment unset')

        try:
            return self.experiment_to_proposal[self.experiment]
        except KeyError:
            # Rare case for debug/daq experiments, roll with it for now
            return self.experiment

    @property
    def run_number(self):
        """Get the run number from the experiment."""
        if self.experiment is None or len(self.experiment) <= 2:
            raise RuntimeError(f'Experiment invalid: {self.experiment}')

        run_number = self.experiment[-2:]
        return f'run{run_number}'

    @functools.lru_cache
    def get_proposal_list(self) -> dict:
        """
        Get the proposal list (a dict, really) for the configured experiment.

        Raises
        ------
        DatabaseError
        """

        run_number = self.run_number
        try:
            logger.debug("Requesting list of proposals in %s", run_number)
            return self._client.getProposalsListForRun(run_number)
        except KeyError as ex:
            # Invalid proposal id for this run
            raise DatabaseError(
                f'Unable to find proposal {self.proposal}'
            ) from ex
        except Exception as ex:
            # Find if our exception gave an HTTP status code and interpret it
            status_code = ex.args[1] if len(ex.args) >= 2 else ''
            if status_code == 500:
                # No information found from run
                reason = f'No run id found for {run_number}'
            elif status_code == 401:
                # Invalid credentials
                reason = 'Invalid credentials'
            else:
                # Unrecognized error
                reason = 'Unable to find run information'
            raise DatabaseError(reason) from ex

    @property
    def beamline(self) -> str:
        """
        Determine the beamline from a proposal + run_number.

        Returns
        -------
        beamline : str
        """

        proposals = self.get_proposal_list()
        return proposals[self.proposal]['Instrument']

    @functools.lru_cache
    def get_run_details(self) -> dict:
        """Get details of the run in a raw dictionary."""
        return self._client.getProposalDetailsForRun(
            self.run_number, self.proposal
        )

    def as_happi_database(self, translations=None) -> dict:
        """
        Based on the current experiment, generate a happi database.

        Parameters
        ----------
        translations : dict, optional
            Translations to use when converting questionnaire items.

        Returns
        -------
        db : dict
            The happi JSON-backend-compatible dictionary.
        """

        return self.to_database(
            beamline=self.beamline,
            run_details=self.get_run_details(),
            translations=translations,
        )

    @staticmethod
    def _translate_items(run_details: dict, table_name: str) -> dict:
        """
        Translate flat questionnaire items into nested dictionaries.

        Parameters
        ----------
        run_details : dict
            The run detail dictionary, from `get_run_details`.

        table_name : str
            The table name (e.g., 'motors' of 'pcdssetup-motors-1-name').

        Returns
        -------
        device_info : dict
        """

        pattern = re.compile(rf'pcdssetup-{table_name}-(\d+)-(\w+)')

        devices = {}
        for field, value in run_details.items():
            match = pattern.match(field)
            if match:
                device_number, name = match.groups()

                if device_number not in devices:
                    devices[device_number] = {}

                # Add the key information to the specific device dictionary
                devices[device_number][name] = value

        return devices

    @staticmethod
    def _create_db_item(info: dict,
                        beamline: str,
                        method_call,
                        ) -> dict:
        """
        Create one database entry given translated questionnaire information.

        Parameters
        ----------
        info : dict
            Device information from `_translate_items`.
        beamline : str
            The beamline with which to associate the entry.
        class_name : str
            The class name to report in the new entry.
        container : str
            The container name to report in the new entry.

        Returns
        -------
        happi_entry : dict
        """

        # Shallow-copy to not modify the original:
        info = dict(info)

        name = info.pop('name', None)
        if not name:
            raise RequiredKeyError('Unable to create an item without a name')

        # There are some easy mistakes we can correct for, otherwise
        # happi validation will fail.

        # 1. A capitalized name:
        name = name.lower()
        entry = method_call(name, beamline, info)

        # Empty strings from the Questionnaire make for invalid entries:
        for key in {'prefix', 'name'}:
            if not entry.get(key):
                raise RequiredKeyError(
                    f"Unable to create an item without key {key}"
                )
        return entry

    @staticmethod
    def to_database(beamline: str,
                    run_details: dict,
                    *,
                    translations: Optional[dict] = None
                    ) -> dict:
        """
        Translate a set of run details into a happi-compatible dictionary.

        Parameters
        ----------
        run_details : dict
            The run detail dictionary, from `get_run_details`.
        beamline : str
            The beamline with which to associate the entry.
        translations : dict, optional
            Translations to use when converting questionnaire items.

        Returns
        -------
        db : dict
            The happi JSON-backend-compatible dictionary.
        """

        happi_db = {}
        if translations is None:
            translations = DEFAULT_TRANSLATIONS

        for table_name, translation in translations.items():
            devices = QuestionnaireHelper._translate_items(
                run_details, table_name)

            if not devices:
                logger.info(
                    "No device information found under '%s'", table_name
                )
                continue

            for device_number, item_info in devices.items():
                logger.debug(
                    '[%s:%s] Found %s', table_name, device_number, item_info
                )
                try:
                    entry = QuestionnaireHelper._create_db_item(
                        info=item_info,
                        beamline=beamline,
                        method_call=translation
                    )
                except RequiredKeyError:
                    logger.debug(
                        'Missing key for %s:%s', table_name, device_number,
                        exc_info=True
                    )
                except Exception as ex:
                    logger.warning(
                        'Failed to create a happi database entry from the '
                        'questionnaire device: %s:%s. %s: %s',
                        table_name, device_number, ex.__class__.__name__, ex,
                    )
                else:
                    identifier = entry['_id']
                    if identifier in happi_db:
                        logger.warning(
                            'Questionnaire name clash: %s (was: %s now: %s)',
                            identifier, happi_db[identifier], entry
                        )
                    happi_db[identifier] = entry

        return happi_db


class QSBackend(JSONBackend):
    """
    Questionniare Backend.

    This backend connects to the LCLS questionnaire and looks at items with the
    key pattern pcds-{}-setup-{}-{}. These fields are then combined and turned
    into proper happi items. The translation of table name to `happi.HappiItem`
    is determined by the :attr:`.translations` dictionary.  The beamline is
    determined by looking where the proposal was submitted.

    Unlike the other backends, this one is read-only. All changes to the device
    information should be done via the web interface. Finally, in order to
    avoid duplicating any code needed to search the device database, the
    `QSBackend` inherits directly from `JSONBackend`. Many of the methods are
    unmodified with exception being that this backend merely searches through
    an in-memory dictionary whereas the `JSONBackend` reads from a file before
    searches.

    Parameters
    ----------
    expname : str
        The experiment name from the elog, e.g. xcslp1915.
    url : str, optional
        Provide a base URL for the Questionnaire. If left as None the
        appropriate URL will be chosen based on your authentication method.
    use_kerberos : bool, optional
        Use a Kerberos ticket to login to the Questionnaire. This is the
        default authentication method.
    user : str, optional
        A username for ws_auth sign-in. If not provided the current login name
        is used.
    pw : str, optional
        A password for ws_auth sign-in. If not provided a password will be
        requested.
    cfg_path : str, optional
        Path to the happi config.
    """

    translations = DEFAULT_TRANSLATIONS

    def __init__(self, expname: str, *, url=None, use_kerberos=True, user=None,
                 pw=None, cfg_path=None):
        # Load cache is unused for this backend, but we have it here for
        # API compatibility with the superclass JSONBackend.
        self._load_cache = None
        # Create our client and gather the raw information from the client
        self._client = QuestionnaireClient(
            url=url, use_kerberos=use_kerberos, user=user, pw=pw
        )

        self.db = self._initialize_database(expname)

    def _initialize_database(self, experiment):
        """Initialize and convert the questionnaire."""
        try:
            self.experiment = experiment
            self.helper = QuestionnaireHelper(self._client)
            self.helper.experiment = experiment
            return self.helper.as_happi_database(
                translations=self.translations
            )
        except Exception:
            logger.error('Failed to load the questionnaire', exc_info=True)
            return {}

    def initialize(self):
        """Can not initialize a new Questionnaire entry from API."""
        raise NotImplementedError("The Questionnaire backend is read-only")

    def load(self):
        """Return the structured dictionary of information."""
        return self.db

    def store(self, *args, **kwargs):
        """The current implementation of this backend is read-only."""
        raise NotImplementedError("The Questionnaire backend is read-only")

    def save(self, *args, **kwargs):
        """The current implementation of this backend is read-only."""
        raise NotImplementedError("The Questionnaire backend is read-only")

    def delete(self, _id):
        """The current implementation of this backend is read-only."""
        raise NotImplementedError("The Questionnaire backend is read-only")



from typing import Any
from unittest.mock import patch

import pytest

from happi import EntryInfo, OphydItem, cache
from happi.item import HappiItem
from happi.loader import fill_template, from_container, load_devices
from happi.utils import create_alias


class TimeDevice(OphydItem):
    days = EntryInfo("Number of days", enforce=int)


def test_fill_template(item: OphydItem):
    # Check that we can properly render a template
    template = "{{name}}"
    assert item.name == fill_template(template, item)
    # Check that we can enforce a type
    template = '{{active}}'
    active = fill_template(template, item, enforce_type=True)
    assert isinstance(active, bool)
    # Check that we will convert a more complex template
    template = '{{name|length()}}'
    text_len = fill_template(template, item, enforce_type=False)
    assert len(item.name) == int(text_len)
    # Check that we can handle non-jinja template
    template = "blah"
    assert template == fill_template(template, item, enforce_type=True)
    # Check that we do not enforce a NoneType
    template = "{{detailed_screen}}"
    assert fill_template(template, item, enforce_type=True) == ""


def test_from_container():
    # Create a datetime item
    d = TimeDevice(name='test', prefix='Tst:This:1', beamline='TST',
                   device_class='datetime.timedelta', args=list(), days=10,
                   kwargs={'days': '{{days}}', 'seconds': 30})
    td = from_container(d)
    # Now import datetime and check that we constructed ours correctly
    import datetime
    assert td == datetime.timedelta(days=10, seconds=30)


def test_caching():
    # Create a datetime item
    d = TimeDevice(name='test', prefix='Tst:This:2', beamline='TST',
                   device_class='types.SimpleNamespace', args=list(), days=10,
                   kwargs={'days': '{{days}}', 'seconds': 30})
    td = from_container(d)
    assert d.name in cache
    assert id(td) == id(from_container(d))
    assert id(td) != id(from_container(d, use_cache=False))
    # Modify md and check we see a reload
    d.days = 12
    assert id(td) != id(from_container(d, use_cache=True))
    # Check with an item where metadata is unavailable
    d = TimeDevice(name='test', prefix='Tst:Delta:3', beamline='TST',
                   device_class='datetime.timedelta', args=list(), days=10,
                   kwargs={'days': '{{days}}', 'seconds': 30})
    td = from_container(d)
    assert id(td) != id(from_container(d))
    assert id(td) != id(from_container(d, use_cache=False))


def test_add_md():
    d = HappiItem(name='test', prefix='Tst:This:3',
                  beamline="TST", args=list(),
                  device_class="happi.HappiItem")
    obj = from_container(d, attach_md=True)
    assert obj.md.extraneous['beamline'] == 'TST'
    assert obj.md.name == 'test'


@pytest.mark.parametrize(
    "threaded",
    [
        pytest.param(False, id="unthreaded"),
        pytest.param(True, id="threaded")
    ],
)
@pytest.mark.parametrize(
    "post_load",
    [
        pytest.param(None, id="no_post_load"),
        pytest.param(lambda x: None, id="post_load"),
    ],
)
@pytest.mark.parametrize(
    "include_load_time",
    [
        pytest.param(False, id="no_load_times"),
        pytest.param(True, id="load_times")
    ],
)
def test_load_devices(threaded: bool, post_load: Any, include_load_time: bool):
    # Create a bunch of devices to load
    devs = [TimeDevice(name='test_1', prefix='Tst1:This', beamline='TST',
                       device_class='datetime.timedelta', args=list(), days=10,
                       kwargs={'days': '{{days}}', 'seconds': 30}),
            TimeDevice(name='test_2', prefix='Tst2:This', beamline='TST',
                       device_class='datetime.timedelta', args=list(), days=10,
                       kwargs={'days': '{{days}}', 'seconds': 30}),
            TimeDevice(name='test_3', prefix='Tst3:This', beamline='TST',
                       device_class='datetime.timedelta', args=list(), days=10,
                       kwargs={'days': '{{days}}', 'seconds': 30}),
            HappiItem(name='bad', prefix='Not:Here', beamline='BAD',
                      device_class='non.existant')]
    # Load our devices
    space = load_devices(
        *devs,
        pprint=True,
        use_cache=False,
        threaded=threaded,
        post_load=post_load,
        include_load_time=include_load_time,
    )
    # Check all our devices are there
    assert all([create_alias(dev.name) in space.__dict__ for dev in devs])
    # Devices were loading properly or exceptions were stored
    import datetime
    assert space.test_1 == datetime.timedelta(days=10, seconds=30)
    assert isinstance(space.bad, ImportError)


def test_filter_kwargs(item_jinja: OphydItem):
    blanks = ['blank_bool', 'blank_list', 'blank_str', 'blank_none']

    # default behavior, allow individuals to decide
    item_jinja._info_attrs['kwargs'].include_default_as_kwarg = True
    dev = from_container(item_jinja, use_cache=False)

    # basic jinja template filling test.
    # only kwargs get attrs in SimpleNamespace
    assert dev.loc == 'LOC'

    # if there is no correspoding EntryInfo, this is a piece of metadata
    # type cannot be matched and value will be returned as string
    assert dev.blank == 'None'
    assert getattr(dev, 'blank_exclude', 'DNE') == 'DNE'
    for bl in blanks:
        assert getattr(dev, bl, 'DNE') == item_jinja._info_attrs[bl].default

    item_jinja._info_attrs['kwargs'].include_default_as_kwarg = False
    filtered_dev = from_container(item_jinja, use_cache=False)

    assert filtered_dev.blank == 'None'
    assert getattr(filtered_dev, 'blank_exclude', 'DNE') == 'DNE'
    for bl in blanks:
        assert getattr(filtered_dev, bl, 'DNE') == 'DNE'


class PostDevice:
    def post_happi_md(self):
        print('post_happi_md hook run')


@pytest.fixture
def item_post_md_hook():
    item = HappiItem(name='post_test',
                     device_class='happi.tests.test_loader.PostDevice')
    return item


def test_post_happi_md(item_post_md_hook: HappiItem):
    with patch('happi.tests.test_loader.PostDevice.post_happi_md') as mock:
        dev = from_container(item_post_md_hook, use_cache=False)
        assert isinstance(dev, PostDevice)
        mock.assert_called_once()

    with patch('happi.tests.test_loader.PostDevice.post_happi_md') as mock:
        dev = from_container(item_post_md_hook, run_post_attach_hook=False,
                             use_cache=False)
        assert isinstance(dev, PostDevice)
        mock.assert_not_called()



import collections.abc
import configparser
import contextlib
import copy
import inspect
import itertools
import logging
import os
import re
import sys
import time as ttime
from collections.abc import Sequence
from typing import Any, Optional

import platformdirs

from . import containers
from .backends import BACKENDS, DEFAULT_BACKEND
from .backends.core import _Backend
from .errors import DatabaseError, EntryError, SearchError, TransferError
from .item import HappiItem
from .loader import from_container

logger = logging.getLogger(__name__)


def _looks_like_database(obj):
    """Returns True if obj is a `Backend` or has certain Backend attributes."""
    return (isinstance(obj, _Backend) or
            all(hasattr(obj, attr) for attr in
                ('find', 'all_items', 'delete', 'save'))
            )


class SearchResult(collections.abc.Mapping):
    """
    A single search result from `Client.search`.

    Examples
    --------
    This result can be keyed for metadata as in:
    >>> result['name']

    The HappiItem can be readily retrieved:
    >>> result.item

    Or the object may be instantiated:
    >>> result.get()

    Attributes
    ----------
    item : happi.HappiItem
        The resulting container.
    metadata : dict
        The resulting HappiItem metadata.
    """

    def __init__(self, client, item):
        self._item = item
        self._instantiated = None
        self.client = client
        self.metadata = item.post()

    @property
    def item(self):
        if self._item is None:
            self._item = self.client.find_item(**self.metadata)
        return self._item

    def get(self, attach_md=True, use_cache=True, threaded=False):
        '''(get) ''' + from_container.__doc__
        if self._instantiated is None:
            self._instantiated = from_container(
                self.item, attach_md=attach_md, use_cache=use_cache,
                threaded=threaded
            )
        return self._instantiated

    def __getitem__(self, item):
        return self.metadata[item]

    def __iter__(self):
        yield from self.metadata

    def __len__(self):
        return len(self.metadata)

    def __repr__(self):
        return (
            f'{self.__class__.__name__}(client={self.client}, '
            f'metadata={self.metadata})'
        )

    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, self.__class__):
            return ((self.client.backend == __o.client.backend)
                    and (self['_id'] == __o['_id']))
        else:
            return False

    def __hash__(self) -> int:
        return hash((self.client.backend, self['_id']))


class Client(collections.abc.Mapping):
    """
    The client to control the contents of the Happi Database.

    Parameters
    ----------
    database : happi.backends.Backend
        A already-instantiated backend.
    kwargs
        Keyword arguments passed to the ``database`` backend.

    Raises
    -----
    DatabaseError
        Raised if the Client fails to instantiate the Database.
    """

    # HappiItem information
    _client_attrs = ['_id', 'type', 'creation', 'last_edit']
    _id_key = 'name'
    _results_wrap_class = SearchResult

    def __init__(self, database=None, **kwargs):
        self._retain_cache = False

        # Use supplied backend
        if database:
            self.backend = database
            if not _looks_like_database(database):
                raise ValueError(f'{database!r} does not look like a database;'
                                 f' expecting an instantiated happi backend')
        # Load database
        else:
            logger.debug("No database given, using '%s'", DEFAULT_BACKEND)
            try:
                self.backend = DEFAULT_BACKEND(**kwargs)
            except Exception as exc:
                raise DatabaseError(
                    f"Failed to instantiate a {DEFAULT_BACKEND} backend"
                ) from exc

    @contextlib.contextmanager
    def retain_cache_context(self, clear_first: bool = True):
        """
        Context manager which can be used to retain the happi backend cache.

        Parameters
        ----------
        clear_first : bool, optional
            Clear the cache before entering the block.  Defaults to True.
        """
        if clear_first:
            self.backend.clear_cache()

        try:
            self._retain_cache = True
            yield
        finally:
            self._retain_cache = False

    def find_document(self, **kwargs):
        """
        Load an item document from the database.

        If multiple matches are found, a single document will be returned to
        the user. How the database will choose to select this item is based
        on each individual implementation

        Parameters
        ----------
        kwargs
            Keyword pairs of information used to locate the item

        Returns
        -------
        document : dict
            A dict that matches the specified information.

        Raises
        ------
        SearchError
            If no document with the given information is found.

        See Also
        --------
        :meth:`.find_item`, :meth:`.search`
        """

        if len(kwargs) == 0:
            raise SearchError("No information pertinent to item given")

        self._maybe_clear_cache()
        matches = list(itertools.islice(self.backend.find(kwargs), 1))
        if not matches:
            raise SearchError(
                "No item information found that matches the search criteria"
            )
        return copy.deepcopy(matches[0])

    def create_item(self, item_cls, **kwargs):
        """
        Instantiate a HappiItem from its container class and keyword arguments.

        Parameters
        ----------
        item_cls : type
            The HappiItem container class to instantiate.

        **kwargs :
            Information to pass through to the item, upon initialization.

        Returns
        -------
        item : object
            An instantiated version of the item.

        Raises
        ------
        TypeError
            If the provided class is not a subclass of :class:`.HappiItem`.

        Example
        -------
        .. code::

            item = client.create_item(Item, name='my_item' ...)
            item = client.create_item('Item', name='my_item',...)
        """

        # string -> class, if in the registry
        if item_cls in containers.registry:
            item_cls = containers.registry[item_cls]

        # Check that this is a valid HappiItem
        if isinstance(item_cls, str):
            raise TypeError(
                f'The container class {item_cls!r} is not in the registry'
            )

        if not (inspect.isclass(item_cls) and issubclass(item_cls, HappiItem)):
            raise TypeError(f"{item_cls!r} is not a subclass of HappiItem")

        item = item_cls(**kwargs)

        def save_item():
            self.add_item(item)

        # Add the method to the item
        item.save = save_item
        return item

    def add_item(self, item):
        """
        Add an item into the database.

        Parameters
        ----------
        item : :class:`.HappiItem`
            The item to store in the database.

        Returns
        -------
        _id : str
            The id of the item added.

        Raises
        ------
        EntryError
            If all of the mandatory information for the item has not been
            specified or there is already an item with that ``id`` in the
            database.
        """

        logger.info("Storing item %r ...", item)
        _id = self._store(item, insert=True)
        logger.info(
            "HappiItem %r has been succesfully added to the database",
            item
        )

        def save_item():
            self._store(item, insert=False)

        item.save = save_item
        return _id

    def _get_item_from_document(self, doc: dict[str, Any]) -> HappiItem:
        """
        Create a HappiItem given its backend-provided document.

        Parameters
        ----------
        doc : dict
            Document from the backend.

        Returns
        -------
        item : :class:`.HappiItem`
            An item that matches the characteristics given.
        """
        try:
            item = self.create_item(doc["type"], **doc)
        except (KeyError, TypeError) as exc:
            raise EntryError(
                "The information relating to the container class "
                "has been modified to the point where the object "
                "can not be initialized, please load the "
                "corresponding document"
            ) from exc

        # Add the save method to the item
        item.save = lambda: self._store(item, insert=False)
        return item

    def find_item(self, **post):
        """
        Query the database for an individual HappiItem.

        If multiple items are found, only the first is returned.

        Parameters
        ----------
        **post :
            Key-value pairs of search criteria used to find the item.

        Raises
        ------
        SearchError
            If no match for the given information is found.

        Returns
        -------
        item : :class:`.HappiItem`
            A HappiItem instance for the document.
        """
        return self._get_item_from_document(self.find_document(**post))

    def load_device(self, use_cache=True, **post):
        """
        Find an item in the database and instantiate it.

        Essentially a shortcut for:

        .. code:: python

            container = client.find_item(name='...')
            item = from_container(container)

        Parameters
        ----------
        post
            Key-value pairs of search criteria passed to
            :meth:`.Client.find_item`.

        use_cache : bool, optional
            Choice to use a cached item. See :meth:`.from_container` for more
            information.

        Returns
        -------
        object
            Instantiated object.
        """
        cntr = self.find_item(**post)
        return from_container(cntr, use_cache=use_cache)

    def change_container(
        self,
        item: HappiItem,
        target: type[HappiItem],
        edits: Optional[dict[str, Any]] = None,
        how: str = 'right'
    ) -> dict[str, Any]:
        """
        Return the kwargs necessary to transfer the information from
        ``item`` into a container ``target``.  Checks are performed to ensure
        the contents are compatible (following enforce requirements in the
        target container)

        This function is built to be a helper function for an interactive
        cli tool, and passes information up accordingly.  If keys are not
        significantly mismatched, this function can be used as is.

        Parameters
        ----------
        item : happi.HappiItem
            HappiItem instance to be transferred to a new container
        target : Type[happi.HappiItem]
            Container to contain new item
        edits : dict[str, Any], optional
            Dictionary of edits to supersede values in original item
        how : str, optional
            Method of resolving the entries between the original item
            and target container.  Can be:
            - "right" : Expect a value for every entry in target container
            - "inner" : Expect values for only entries in BOTH original
            item and target container

        Raises
        ------
        TransferError
            If there is a problem tranferring item into container

        Returns
        -------
        new_kwargs : Dict[str, Any]
            kwargs to pass into a new HappiItem
        """
        edits = edits or {}

        # grab all keys, extraneous and otherwise
        item_post = item.post()
        if how == 'right':
            # Try to fill every value in target
            base_names = target.info_names
        elif how == 'inner':
            # only keys in both original and target
            base_names = [n for n in target.info_names
                          if n in item_post.keys()]
        else:
            raise ValueError(f'Improper merge method: {how}')

        target_entries = {e.key: e for e in target.entry_info}
        new_kwargs = {}
        new_kwargs.update(edits)  # can contain extraneous data
        for name in base_names:
            try:
                # validate every value being placed into target
                # grab edit value, defaulting to item value
                old_val = new_kwargs.get(name, item_post.get(name))
                # no value found in either, continue and handle later
                if old_val is None:
                    continue
                val = target_entries[name].enforce_value(old_val)

                new_kwargs.update({name: val})
            except ValueError as e:
                # Repackage exception with key information
                raise TransferError(e, name)

        # ensure all mandatory info filled
        for req_name in target.mandatory_info:
            if req_name not in new_kwargs:
                msg = f'mandatory field {req_name} missing a value'
                raise TransferError(msg, req_name)

        return new_kwargs

    def validate(self):
        """
        Validate all of the items in the database.

        This is done by attempting to initialize each item and asserting
        their mandatory information is present. Information is written to the
        logger.

        Returns
        -------
        ids : list of str
            List of item ids that have failed verification.
        """

        bad = list()
        logger.debug('Loading database to validate contained devices ...')
        for doc in self.backend.all_items:
            # Try and load device based on database info
            _id = doc.get(self._id_key, "(unknown id)")
            try:
                logger.debug("Attempting to initialize %s...", _id)
                item = self._get_item_from_document(doc)
                logger.debug("Attempting to validate ...")
                self._validate_item(item)
            except Exception as e:
                logger.warning("Failed to validate %s because %s", _id, e)
                bad.append(_id)
            else:
                logger.debug('Successfully validated %s', _id)
        return bad

    @property
    def all_items(self):
        """A list of all contained items."""
        return [res.item for res in self.search()]

    def __getitem__(self, key):
        """Get an item ID."""
        try:
            item = self._get_item_from_document(self.backend.get_by_id(key))
        except Exception as ex:
            raise KeyError(key) from ex

        return SearchResult(client=self, item=item)

    def __iter__(self):
        for info in self.backend.find({}):
            yield info['_id']

    def __len__(self):
        return len(self.all_items)

    def _maybe_clear_cache(self):
        """Clear the backend cache if not in a retain-cache block."""
        if not self._retain_cache:
            self.backend.clear_cache()

    def _get_search_results(
        self, docs: Sequence[dict[str, Any]], *, wrap_cls: Optional[type] = None
    ) -> list[SearchResult]:
        """
        Return search results to the user, optionally wrapping with a class.
        """
        wrap_cls = wrap_cls or self._results_wrap_class
        assert wrap_cls is not None

        results = []
        for doc in docs:
            try:
                results.append(
                    wrap_cls(
                        client=self,
                        item=self._get_item_from_document(doc)
                    )
                )
            except Exception as exc:
                logger.warning(
                    "Entry for %s is malformed (%s). Skipping.",
                    doc["name"], exc
                )
        return results

    def search_range(
        self, key: str, start: float, end: Optional[float] = None, **kwargs
    ) -> list[SearchResult]:
        """
        Search the database for an item or items using a numerical range.

        Parameters
        -----------
        key : str
            Database key to search.
        start : float, optional
            Minimum beamline position to include items.
        end : float, optional
            Maximum beamline position to include items.
        **kwargs : optional
            Additional information used to filter through the database
            structured as key-value pairs for the desired pieces of EntryInfo.

        Returns
        -------
        list of SearchResult

        Example
        -------
        .. code::

            gate_valves = client.search_range('z', 0, 100, type='Valve')
            hxr_valves  = client.search_range('z', 0, 100, type='Valve',
                                              beamline='HXR')
        """

        self._maybe_clear_cache()
        items = self.backend.find_range(key, start=start, stop=end,
                                        to_match=kwargs)
        return self._get_search_results(items)

    def search(self, **kwargs):
        """
        Search the database for item(s).

        Parameters
        -----------
        **kwargs
            Information to filter through the database structured as key, value
            pairs for the desired pieces of EntryInfo.

        Returns
        -------
        res : list of SearchResult
            The search results.

        Example
        -------
        .. code::

            gate_valves = client.search(type='Valve')
            hxr_valves  = client.search(type='Valve', beamline='HXR')
        """

        self._maybe_clear_cache()
        items = self.backend.find(kwargs)
        return self._get_search_results(items)

    def search_regex(
        self, flags=re.IGNORECASE, **kwargs
    ) -> list[SearchResult]:
        """
        Search the database for items(s).

        Parameters
        -----------
        flags : int, optional
            Defaulting to ``re.IGNORECASE``, these flags are used for the
            regular expressions passed in.
        **kwargs
            Information to filter through the database structured as key, value
            pairs for the desired pieces of EntryInfo.  Every value is allowed
            to contain a Python regular expression.

        Returns
        -------
        list of SearchResult

        Example
        -------
        .. code::

            gate_valves = client.search_regex(beamline='Valve.*')
            three_valves = client.search_regex(_id='VALVE[123]')
        """

        self._maybe_clear_cache()
        items = self.backend.find_regex(kwargs, flags=flags)
        return self._get_search_results(items)

    def export(self, path=sys.stdout, sep='\t', attrs=None):
        """
        Export the contents of the database into a text file.

        Parameters
        ----------
        path : File Handle
            File-like object to save text file. Defaults to stdout.
        sep : str
            Separator to place inbetween columns of information.
        attrs : iterable
            Attributes to include, these will be a list of values.
        """

        # Load documents
        devs = self.all_items
        logger.info('Creating file at %s ...', path)
        # Load item information
        with path as f:
            for dev in devs:
                try:
                    f.write(sep.join([getattr(dev, attr)
                                      for attr in attrs])
                            + '\n')
                except KeyError as e:
                    logger.error("HappiItem %s was missing attribute %s",
                                 dev.name, e)

    def remove_item(self, item):
        """
        Remove an item from the database.

        Parameters
        ----------
        item : :class:`.HappiItem`
            HappiItem to be removed from the database.
        """

        # HappiItem Check
        if not isinstance(item, HappiItem):
            raise ValueError("Must supply an object of type `HappiItem`")
        logger.info("Attempting to remove %r from the "
                    "collection ...", item)
        # Check that item is in the database
        _id = getattr(item, self._id_key)
        self.backend.delete(_id)

    def _validate_item(self, item):
        """Validate that an item has all of the mandatory information."""
        logger.debug('Validating item %r ...', item)

        # Check type
        if not isinstance(item, HappiItem):
            raise ValueError(f"{item!r} is not a subclass of HappiItem")
        logger.debug('Checking mandatory information has been entered ...')
        # Check that all mandatory info has been entered
        missing = [info.key for info in item.entry_info
                   if not info.optional and
                   info.default == getattr(item, info.key)]
        # Abort initialization if missing mandatory info
        if missing:
            raise EntryError(
                "Missing mandatory information ({}) for {}"
                "".format(", ".join(missing), item.__class__.__name__)
            )
        logger.debug('HappiItem %r has been validated.', item)

    def _store(self, item, insert=False):
        """
        Store a document in the database.

        Parameters
        ----------
        item : :class:`.HappiItem`
            HappiItem to save.
        insert : bool, optional
            Set to `True` if this is a new entry.

        Raises
        ------
        DuplicateError
            When ``_id`` already exists.
        EntryError
            If the item doesn't the correct information.

        Todo
        ----
        Enforce parent is an already entered name
        """

        logger.debug('Loading an item into the collection ...')

        # Validate item is ready for storage
        self._validate_item(item)
        # Grab information from item
        post = item.post()
        # save the old name in case the user is trying to
        # change the 'name' of an entry
        the_old_name = post.get('_id', None)
        # Store creation time
        creation = post.get('creation', ttime.ctime())
        # Clean supplied information
        [post.pop(key) for key in self._client_attrs if key in post]
        # Note that item has some unrecognized metadata
        for key in post.keys():
            if key not in item.info_names:
                logger.debug(
                    "HappiItem %r defines an extra piece of "
                    "information under the keyword %s",
                    item,
                    key,
                )
        # Add metadata from the Client Side

        tpe = containers.registry.entry_for_class(item.__class__)
        post.update({'type': tpe,
                     'creation': creation,
                     'last_edit': ttime.ctime()})

        # Find id
        try:
            _id = post[self._id_key]
        except KeyError:
            raise EntryError('HappiItem did not supply the proper information '
                             'to interface with the database, missing {}'
                             ''.format(self._id_key))
        # In case we want to update the name of an entry
        # We want to add a new entry, and delete the old one
        if the_old_name and the_old_name != post[self._id_key]:
            # Store information for the new entry
            logger.info('Saving new entry %s ...', _id)
            self.backend.save(_id, post, insert=True)
            # Remove the information for the old entry
            logger.info('Removing old entry %s ...', the_old_name)
            self.backend.delete(the_old_name)
        else:
            # Store information
            logger.info('Adding / Modifying information for %s ...', _id)
            self.backend.save(_id, post, insert=insert)
        return _id

    @classmethod
    def from_config(cls, cfg=None):
        """
        Create a client from a configuration file specification.

        Configuration files looking something along the lines of:

        .. code::

            [DEFAULT]
            path=path/to/my/db.json

        All key value pairs will be passed directly into backend construction
        with the exception of the key ``backend`` which can be used to specify
        a specific type of backend if this differs from the configured default.

        Parameters
        ----------
        cfg : str, optional
            Path to a configuration file. If not entered, :meth:`.find_config`
            will be use.
        """

        # Find a configuration file
        if not cfg:
            cfg = cls.find_config()

        if not os.path.exists(cfg):
            raise RuntimeError(f'happi configuration file not found: {cfg!r}')

        # Parse configuration file
        cfg_parser = configparser.ConfigParser()
        cfg_file = cfg_parser.read(cfg)
        logger.debug("Loading configuration file at %r", cfg_file)

        # Gather the backend from each section
        sub_backends = []
        for section in cfg_parser.keys():
            bknd = cls._get_backend_from_config(cfg_parser[section], cfg)
            if bknd:
                sub_backends.append(bknd)

        if len(sub_backends) == 0:
            raise RuntimeError('No valid backends loaded')
        elif len(sub_backends) == 1:
            # return single backend client
            return cls(database=sub_backends[0])
        else:
            # multi backend configuration have multiple sections
            multi_backend = BACKENDS['multi']

            # move DEFAULT from index 0 to -1, if it exists
            if dict(cfg_parser['DEFAULT']):
                default_backend = sub_backends.pop(0)
                sub_backends.append(default_backend)

            # return client with multi_backend
            db = multi_backend(backends=sub_backends)
            return cls(database=db)

    @staticmethod
    def _get_backend_from_config(
        cfg_section: configparser.SectionProxy,
        cfg: str
    ) -> _Backend:
        # If a backend is specified use it, otherwise default
        if 'backend' in cfg_section:
            db_str = cfg_section.pop('backend')
            try:
                backend = BACKENDS[db_str]
            except KeyError:
                raise RuntimeError(
                    f'Happi backend {db_str!r} unavailable'
                ) from None
        else:
            backend = DEFAULT_BACKEND

        logger.debug(
            "Using Happi backend %r with kwargs %s",
            backend, cfg_section
        )

        if not dict(cfg_section):
            # section is empty, skip
            return

        # Create our database with provided kwargs
        try:
            database = backend(**cfg_section, cfg_path=cfg)
            return database
        except Exception as ex:
            raise RuntimeError(
                f'Unable to instantiate the client. Please verify that '
                f'your HAPPI_CFG points to the correct file and has '
                f'the required configuration settings. In {cfg!r}, found '
                f'settings: {dict(cfg_section)}.'
            ) from ex

    @staticmethod
    def find_config():
        """
        Search for a ``happi`` configuration file.

        We first query the environment variable ``$HAPPI_CFG`` to see if this
        points to a specific configuration file. If this is not present, the
        variable set by ``$XDG_CONFIG_HOME`` or if  that is not set
        ``~/.config``

        Returns
        -------
        path : str
            Absolute path to configuration file.

        Raises
        ------
        EnvironmentError
            If no configuration file can be found by the methodology detailed
            above.
        """

        # Point to with an environment variable
        if os.environ.get('HAPPI_CFG', False):
            happi_cfg = os.environ.get('HAPPI_CFG')
            logger.debug("Found $HAPPI_CFG specification for Client "
                         "configuration at %s", happi_cfg)
            return happi_cfg
        # Search in the current directory and home directory
        else:
            config_dirs = [os.environ.get('XDG_CONFIG_HOME', "."),
                           os.path.expanduser('~/.config'),
                           platformdirs.user_config_dir("happi")]
            for directory in config_dirs:
                logger.debug('Searching for Happi config in %s', directory)
                for path in ('.happi.cfg', 'happi.cfg'):
                    full_path = os.path.join(directory, path)

                    if os.path.exists(full_path):
                        logger.debug("Found configuration file at %r", full_path)
                        return full_path
        # If found nothing
        raise OSError("No happi configuration file found. Check HAPPI_CFG.")

    def choices_for_field(self, field):
        """
        List all choices for a given field.

        Parameters
        ----------
        field : string
            Field name to list all possible choices for
            i.e 'beamline', 'name', 'z', 'prefix', etc.

        Raises
        ------
        SearchError
            If no items in the database have an entry for the given field.

        Returns
        -------
        field_choices : list
            List of choices for a given field that are in the database.
        """

        field_choices = set()
        for dev in self.all_items:
            try:  # Want to ignore error if 'dev' doesn't have 'field'
                choice = getattr(dev, field)
                field_choices.add(choice)
            except AttributeError:
                pass
        if len(field_choices) == 0:
            raise SearchError('No entries found with given field')
        return field_choices



import re

import pytest
from click.testing import CliRunner

import happi
from happi.cli import happi_cli


@pytest.fixture(scope='function')
def client(bad_happi_cfg: str):
    """ misconfigured database """
    return happi.client.Client.from_config(cfg=bad_happi_cfg)


def number_failed_devices(output: str):
    """ Parse cli output for number of failed devices """
    summary_line = [line for line in output.split('\n')
                    if '# devices failed' in line][0]

    match = re.search(r'(\d*) / (\d*)', summary_line)
    return int(match[1])


@pytest.mark.parametrize("n_fails, check", [
    (1, "check_name_match_id"),
    (4, "check_instantiation"),  # simplenamespace does not take args
    (1, "check_extra_info"),
    (2, "check_args_kwargs_match"),
    (1, "check_unfilled_mandatory_info")
])
def test_audit_cli(
    runner: CliRunner,
    bad_happi_cfg: str,
    n_fails: int,
    check: str
):
    res = runner.invoke(happi_cli,
                        ['--path', bad_happi_cfg, 'audit', '-c', check, '*'])
    # check that device failed
    print(res.output)
    assert number_failed_devices(res.output) == n_fails



from .version import __version__  # noqa: F401

__all__ = [
    "Client",
    "EntryInfo",
    "HappiItem",
    "OphydItem",
    "SearchResult",
    "cache",
    "from_container",
    "load_devices",
]
from .client import Client, SearchResult
from .item import EntryInfo, HappiItem, OphydItem
from .loader import cache, from_container, load_devices
